from collections import defaultdict
from typing import TYPE_CHECKING, Any, Optional

from edgy.exceptions import QuerySetError

if TYPE_CHECKING:
    from edgy import Model, QuerySet
    from edgy.core.db.models.types import BaseModelType


class Prefetch:
    """
    Class object that allows the prefetching
    of specific fields.
    """

    def __init__(
        self,
        related_name: str,
        to_attr: str,
        queryset: Optional["QuerySet"] = None,
    ) -> None:
        self.related_name = related_name
        self.to_attr = to_attr
        self.queryset: Optional[QuerySet] = queryset
        self._is_finished = False
        self._bake_prefix: str = ""
        self._baked_results: dict[tuple[str, ...], list[Any]] = defaultdict(list)
        self._baked = False

    async def init_bake(self, model_class: type["Model"]) -> None:
        if self._baked or not self._is_finished or self.queryset is None:
            return
        self._baked = True
        # we want all at once and async iterate
        async for result in self.queryset._execute_iterate(True):
            # a bit hacky but we need the current row
            model_key = model_class.create_model_key_from_sqla_row(
                self.queryset._cache_current_row, row_prefix=self._bake_prefix
            )
            self._baked_results[model_key].append(result)


def check_prefetch_collision(model: "BaseModelType", related: Prefetch) -> Prefetch:
    if (
        hasattr(model, related.to_attr)
        or related.to_attr in model.meta.fields
        or related.to_attr in model.meta.managers
    ):
        raise QuerySetError(
            f"Conflicting attribute to_attr='{related.related_name}' with '{related.to_attr}' in {model.__class__.__name__}"
        )
    return related


class PrefetchMixin:
    """
    Query used to perform a prefetch_related into the models and
    subsequent queries.
    """

    def prefetch_related(self, *prefetch: Prefetch) -> "QuerySet":
        """
        Performs a reverse lookup for the foreignkeys. This is different
        from the select_related. Select related performs a follows the SQL foreign relation
        whereas the prefetch_related follows the python relations.
        """
        queryset: QuerySet = self._clone()

        if any(not isinstance(value, Prefetch) for value in prefetch):
            raise QuerySetError("The prefetch_related must have Prefetch type objects only.")

        queryset._prefetch_related = [*self._prefetch_related, *prefetch]
        return queryset



import os
import sys
import typing
from dataclasses import dataclass
from importlib import import_module
from pathlib import Path

from edgy.cli.constants import (
    DISCOVERY_FILES,
    DISCOVERY_FUNCTIONS,
    EDGY_DB,
    EDGY_DISCOVER_APP,
    EDGY_EXTRA,
)
from edgy.exceptions import CommandEnvironmentError


@dataclass
class Scaffold:
    """
    Simple Application scaffold that holds the
    information about the app and the path to
    the same app.
    """

    path: str
    app: typing.Any


@dataclass
class MigrationEnv:
    """
    Loads an arbitraty application into the object
    and returns the App.
    """

    path: typing.Optional[str] = None
    app: typing.Optional[typing.Any] = None
    command_path: typing.Optional[str] = None

    def load_from_env(
        self, path: typing.Optional[str] = None, enable_logging: bool = True
    ) -> "MigrationEnv":
        """
        Loads the environment variables into the scaffold.
        """
        # Adds the current path where the command is being invoked
        # To the system path
        cwd = Path().cwd()
        command_path = str(cwd)
        if command_path not in sys.path:
            sys.path.append(command_path)
        try:
            import dotenv

            dotenv.load_dotenv()
        except ImportError:
            ...

        _path = path if path else os.getenv(EDGY_DISCOVER_APP)
        _app = self.find_app(path=_path, cwd=cwd)

        return MigrationEnv(path=_app.path, app=_app.app)

    def import_app_from_string(cls, path: typing.Optional[str] = None) -> Scaffold:
        if path is None:
            raise CommandEnvironmentError(
                detail="Path cannot be None. Set env `EDGY_DEFAULT_APP` or use `--app` instead."
            )
        module_str_path, app_name = path.split(":")
        module = import_module(module_str_path)
        app = getattr(module, app_name)
        return Scaffold(path=path, app=app)

    def _find_app_in_folder(  # type: ignore
        self, path: Path, cwd: Path
    ) -> typing.Union[typing.Callable[..., typing.Any], Scaffold, None]:
        """
        Iterates inside the folder and looks up to the DISCOVERY_FILES.
        """
        for discovery_file in DISCOVERY_FILES:
            file_path = path / discovery_file
            if not file_path.exists():
                continue

            dotted_path = ".".join(file_path.relative_to(cwd).with_suffix("").parts)

            # Load file from module
            module = import_module(dotted_path)

            # Iterates through the elements of the module.
            for attr, value in module.__dict__.items():
                if callable(value) and (hasattr(value, EDGY_DB) or hasattr(value, EDGY_EXTRA)):
                    app_path = f"{dotted_path}:{attr}"
                    return Scaffold(app=value, path=app_path)

            # Iterate over default pattern application functions
            for func in DISCOVERY_FUNCTIONS:
                if hasattr(module, func):
                    app_path = f"{dotted_path}:{func}"
                    fn = getattr(module, func)()

                    if hasattr(fn, EDGY_DB) or hasattr(fn, EDGY_EXTRA):
                        return Scaffold(app=fn, path=app_path)

    def find_app(self, path: typing.Optional[str], cwd: Path) -> Scaffold:
        """
        Loads the application based on the path provided via env var.

        If no --app is provided, goes into auto discovery up to one level.
        """

        if path:
            return self.import_app_from_string(path)

        scaffold: typing.Optional[Scaffold] = None

        # Check current folder
        scaffold = self._find_app_in_folder(cwd, cwd)  # type: ignore
        if scaffold:
            return scaffold

        # Goes into auto discovery mode for one level, only.

        for folder in cwd.iterdir():
            folder_path = cwd / folder
            scaffold = self._find_app_in_folder(folder_path, cwd)  # type: ignore

            if not scaffold:
                continue
            break

        if not scaffold:
            raise CommandEnvironmentError(detail="Could not find Edgy in any application.")
        return scaffold



import asyncio
import weakref
from collections.abc import Awaitable
from contextvars import copy_context
from threading import Event, Thread
from typing import Any, Optional, cast


async def _coro_helper(awaitable: Awaitable, timeout: Optional[float]) -> Any:
    if timeout is not None and timeout > 0:
        return await asyncio.wait_for(awaitable, timeout)
    return await awaitable


# python <= 3.8 doesn't support WeakKeyDictionary as type
weak_subloop_map: dict[asyncio.AbstractEventLoop, asyncio.AbstractEventLoop] = cast(
    dict[asyncio.AbstractEventLoop, asyncio.AbstractEventLoop], weakref.WeakKeyDictionary()
)


async def _startup(old_loop: asyncio.AbstractEventLoop, is_initialized: Event) -> None:
    new_loop = asyncio.get_running_loop()
    weakref.finalize(old_loop, new_loop.stop)
    weak_subloop_map[old_loop] = new_loop
    is_initialized.set()


def _init_thread(old_loop: asyncio.AbstractEventLoop, is_initialized: Event) -> None:
    loop = asyncio.new_event_loop()
    # keep reference
    task = loop.create_task(_startup(old_loop, is_initialized))
    try:
        try:
            loop.run_forever()
        except RuntimeError:
            pass
        finally:
            # now all inits wait
            is_initialized.clear()
        loop.run_until_complete(loop.shutdown_asyncgens())
    finally:
        del task
        loop.close()


def get_subloop(loop: asyncio.AbstractEventLoop) -> asyncio.AbstractEventLoop:
    sub_loop = weak_subloop_map.get(loop)
    if sub_loop is None:
        is_initialized = Event()
        thread = Thread(target=_init_thread, args=[loop, is_initialized], daemon=True)
        thread.start()
        is_initialized.wait()
        return weak_subloop_map[loop]

    return sub_loop


def run_sync(awaitable: Awaitable, timeout: Optional[float] = None) -> Any:
    """
    Runs the queries in sync mode
    """
    loop = None
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        loop = None

    if loop is None:
        return asyncio.run(_coro_helper(awaitable, timeout))
    else:
        ctx = copy_context()
        # the context of the coro seems to be switched correctly
        # in case of problems, we can switch to threadexecutors with asyncio.run but this is not as performant
        return asyncio.run_coroutine_threadsafe(
            ctx.run(_coro_helper, awaitable, timeout), get_subloop(loop)
        ).result()



from typing import Any, Optional, cast

from edgy.core.db.models.metaclasses import (
    BaseModelMeta,
    MetaInfo,
)


def _check_model_inherited_tenancy(bases: tuple[type, ...]) -> bool:
    for base in bases:
        meta: Optional[MetaInfo] = getattr(base, "meta", None)
        if meta is None:
            continue
        if hasattr(meta, "is_tenant"):
            return cast(bool, meta.is_tenant)

    return False


class TenantMeta(MetaInfo):
    __slots__ = ("is_tenant", "register_default")
    register_default: bool

    def __init__(self, meta: Any = None, **kwargs: Any) -> None:
        self.is_tenant = getattr(meta, "is_tenant", None)
        # register in models too, so there is a default
        # otherwise they only exist as tenant models
        self.register_default: bool = getattr(meta, "register_default", True)

        # must be last
        super().__init__(meta, **kwargs)

    def set_tenant(self, is_tenant: bool) -> None:
        self.is_tenant = is_tenant


class BaseTenantMeta(BaseModelMeta):
    """
    The metaclass for the base tenant used by the Edgy contrib mixin.

    This should only be used by the contrib or if you decided to use
    your own tenant model using the `is_tenant` inside the `Meta` object.
    """

    def __new__(cls, name: str, bases: tuple[type, ...], attrs: Any, **kwargs: Any) -> Any:
        new_model = super().__new__(cls, name, bases, attrs, meta_info_class=TenantMeta, **kwargs)
        if new_model.meta.is_tenant is None:
            new_model.meta.is_tenant = _check_model_inherited_tenancy(bases)

        if (
            new_model.meta.registry
            and new_model.meta.is_tenant
            and not new_model.meta.abstract
            and not new_model.__is_proxy_model__
        ):
            assert (
                new_model.__reflected__ is False
            ), "Reflected models are not compatible with multi_tenancy"

            if not new_model.meta.register_default:
                # remove from models
                new_model.meta.registry.models.pop(new_model.__name__, None)
            new_model.meta.registry.tenant_models[new_model.__name__] = new_model
        return new_model



import ipaddress
from abc import abstractmethod
from typing import Any

import sqlalchemy


class BaseFieldProtocol(sqlalchemy.TypeDecorator):
    """
    When implementing a field representation from SQLAlchemy, the protocol will be enforced
    """

    impl: Any
    cache_ok: bool

    @abstractmethod
    def load_dialect_impl(self, dialect: Any) -> Any:
        raise NotImplementedError("load_dialect_impl must be implemented")

    @abstractmethod
    def process_bind_param(self, value: Any, dialect: Any) -> Any:
        raise NotImplementedError("process_bind_param must be implemented")

    @abstractmethod
    def process_result_value(self, value: Any, dialect: Any) -> Any:
        """
        Processes the value coming from the database in a column-row style.
        """
        raise NotImplementedError("process_result_value must be implemented")


class IPAddress(BaseFieldProtocol):
    impl: str = sqlalchemy.CHAR  # type: ignore
    cache_ok: bool = True

    def load_dialect_impl(self, dialect: Any) -> Any:
        if dialect.name not in {"postgres", "postgresql"}:
            return dialect.type_descriptor(sqlalchemy.CHAR(45))
        return dialect.type_descriptor(sqlalchemy.dialects.postgresql.INET())

    def process_bind_param(self, value: Any, dialect: Any) -> Any:
        if value is not None:
            return str(value)

    def process_result_value(self, value: Any, dialect: Any) -> Any:
        if value is None:
            return value
        if not isinstance(value, (ipaddress.IPv4Address, ipaddress.IPv6Address)):
            value = ipaddress.ip_address(value)
        return value



import copy
import inspect
from collections.abc import Sequence
from typing import (
    TYPE_CHECKING,
    Any,
    Optional,
    Union,
    cast,
)

from pydantic import BaseModel

from edgy.core.db.constants import ConditionalRedirect
from edgy.core.db.context_vars import MODEL_GETATTR_BEHAVIOR
from edgy.core.db.fields.base import BaseCompositeField
from edgy.core.db.fields.core import FieldFactory
from edgy.core.db.fields.types import BaseFieldType
from edgy.exceptions import FieldDefinitionError

if TYPE_CHECKING:
    from edgy.core.db.models.types import BaseModelType


class ConcreteCompositeField(BaseCompositeField):
    """
    Conrete, internal implementation of the CompositeField
    """

    def __init__(
        self,
        *,
        inner_fields: Union[
            Sequence[Union[str, tuple[str, BaseFieldType]]],
            type["BaseModelType"],
            dict[str, BaseFieldType],
        ] = (),
        **kwargs: Any,
    ):
        self.inner_field_names: list[str] = []
        self.embedded_field_defs: dict[str, BaseFieldType] = {}
        if hasattr(inner_fields, "meta"):
            kwargs.setdefault("model", inner_fields)
            kwargs.setdefault("inherit", inner_fields.meta.inherit)
            inner_fields = inner_fields.meta.fields
        if isinstance(inner_fields, dict):
            inner_fields = inner_fields.items()  # type: ignore
        owner = kwargs.pop("owner", None)
        self.prefix_embedded: str = kwargs.pop("prefix_embedded", "")
        self.unsafe_json_serialization: bool = kwargs.pop("unsafe_json_serialization", False)
        self.absorb_existing_fields: bool = kwargs.pop("absorb_existing_fields", False)
        self.model: Optional[Union[type[BaseModel], type[ConditionalRedirect]]] = kwargs.pop(
            "model", None
        )
        if self.model is not None and issubclass(self.model, BaseModel):
            kwargs["field_type"] = self.model
            kwargs["annotation"] = self.model
        for field in inner_fields:
            if isinstance(field, str):
                self.inner_field_names.append(field)
            elif field[1].inherit:
                # don't copy non inherit fields for excluding PKField and other surprises
                field_name = field[0]
                # for preventing suddenly invalid field names
                if self.prefix_embedded.endswith("_") and field_name.startswith("_"):
                    raise FieldDefinitionError(
                        f"_ prefixed fields are not supported: {field_name} with prefix ending with _"
                    )
                field_name = f"{self.prefix_embedded}{field_name}"
                # set field_name and owner
                # parent is None because not fully initialized
                field_def = field[1].embed_field(
                    self.prefix_embedded, field_name, owner=owner, parent=None
                )
                if field_def is not None:
                    field_def.exclude = True
                    self.inner_field_names.append(field_def.name)
                    self.embedded_field_defs[field_def.name] = field_def

        return super().__init__(
            owner=owner,
            # this is just a holder for real fields
            null=True,
            **kwargs,
        )

    def translate_name(self, name: str) -> str:
        if self.prefix_embedded and name in self.embedded_field_defs:
            # PYTHON 3.8 compatibility
            return name.removeprefix(self.prefix_embedded)
        return name

    def embed_field(
        self,
        prefix: str,
        new_fieldname: str,
        owner: Optional[type["BaseModelType"]] = None,
        parent: Optional[BaseFieldType] = None,
    ) -> BaseFieldType:
        field_copy = cast(
            BaseFieldType, super().embed_field(prefix, new_fieldname, owner=owner, parent=self)
        )
        field_copy.prefix_embedded = f"{prefix}{field_copy.prefix_embedded}"
        embedded_field_defs = field_copy.embedded_field_defs
        field_copy.inner_field_names = [
            f"{prefix}{field_name}"
            for field_name in field_copy.inner_field_names
            if field_name not in embedded_field_defs
        ]
        field_copy.embedded_field_defs = {}
        for field_name, field in embedded_field_defs.items():
            # for preventing suddenly invalid field names
            if self.prefix_embedded.endswith("_") and field_name.startswith("_"):
                raise FieldDefinitionError(
                    f"_ prefixed fields are not supported: {field_name} with prefix ending with _"
                )
            field_name = f"{prefix}{field_name}"
            field_def = field.embed_field(prefix, field_name, owner)
            if field_def is not None:
                field_def.exclude = True
                field_copy.inner_field_names.append(field_def.name)
                field_copy.embedded_field_defs[field_def.name] = field_def
        return field_copy

    async def aget(
        self, instance: "BaseModelType", owner: Any = None
    ) -> Union[dict[str, Any], Any]:
        d = {}
        token = MODEL_GETATTR_BEHAVIOR.set("coro")
        try:
            for key in self.inner_field_names:
                translated_name = self.translate_name(key)
                value = getattr(instance, key)
                if inspect.isawaitable(value):
                    value = await value
                d[translated_name] = value
        finally:
            MODEL_GETATTR_BEHAVIOR.reset(token)
        if self.model is not None and self.model is not ConditionalRedirect:
            return self.model(**d)
        return d

    def __get__(self, instance: "BaseModelType", owner: Any = None) -> Union[dict[str, Any], Any]:
        assert len(self.inner_field_names) >= 1
        if self.model is ConditionalRedirect and len(self.inner_field_names) == 1:
            try:
                return getattr(instance, self.inner_field_names[0])
            except AttributeError:
                if not instance._loaded_or_deleted:
                    raise AttributeError("not loaded") from None
                return None
        if MODEL_GETATTR_BEHAVIOR.get() == "coro":
            return self.aget(instance, owner=owner)
        d = {}
        for key in self.inner_field_names:
            translated_name = self.translate_name(key)
            try:
                d[translated_name] = getattr(instance, key)
            except (AttributeError, KeyError):
                if not instance._loaded_or_deleted:
                    raise AttributeError("not loaded") from None
                pass
        if self.model is not None and self.model is not ConditionalRedirect:
            return self.model(**d)
        return d

    def clean(self, field_name: str, value: Any, for_query: bool = False) -> dict[str, Any]:
        assert len(self.inner_field_names) >= 1
        if (
            self.model is ConditionalRedirect
            and len(self.inner_field_names) == 1
            # we first only redirect both
            and not isinstance(value, (dict, BaseModel))
        ):
            field = self.owner.meta.fields[self.inner_field_names[0]]
            return field.clean(self.inner_field_names[0], value, for_query=for_query)
        return super().clean(field_name, value, for_query=for_query)

    def to_model(
        self,
        field_name: str,
        value: Any,
    ) -> dict[str, Any]:
        assert len(self.inner_field_names) >= 1
        if (
            self.model is ConditionalRedirect
            and len(self.inner_field_names) == 1
            # we first only redirect both
            and not isinstance(value, (dict, BaseModel))
        ):
            field = self.owner.meta.fields[self.inner_field_names[0]]
            return field.to_model(self.inner_field_names[0], value)
        return super().to_model(field_name, value)

    def get_embedded_fields(
        self, name: str, fields: dict[str, "BaseFieldType"]
    ) -> dict[str, "BaseFieldType"]:
        retdict = {}
        # owner is set: further down in hierarchy, or uninitialized embeddable, where the owner = model
        # owner is not set: current class

        if not self.absorb_existing_fields:
            if self.owner is None:
                duplicate_fields = set(self.embedded_field_defs.keys()).intersection(
                    {k for k, v in fields.items() if v.owner is None}
                )
                if duplicate_fields:
                    raise ValueError(f"duplicate fields: {', '.join(duplicate_fields)}")
            for field_name, field in self.embedded_field_defs.items():
                existing_field = fields.get(field_name)
                if (
                    existing_field is not None
                    and existing_field.owner is None
                    and self.owner is not None
                ):
                    continue
                # now there should be no collisions anymore
                cloned_field = copy.copy(field)
                # set to the current owner of this field, required in collision checks
                cloned_field.owner = self.owner
                cloned_field.inherit = False
                retdict[field_name] = cloned_field
            return retdict
        for field_name, field in self.embedded_field_defs.items():
            if field_name not in fields:
                cloned_field = copy.copy(field)
                # set to the current owner of this field, required in collision checks
                cloned_field.owner = self.owner
                cloned_field.inherit = False
                retdict[field_name] = cloned_field
            else:
                absorbed_field = fields[field_name]
                if not getattr(absorbed_field, "skip_absorption_check", False) and not issubclass(
                    absorbed_field.field_type, field.field_type
                ):
                    raise ValueError(
                        f'absorption failed: field "{field_name}" handle the type: {absorbed_field.field_type}, required: {field.field_type}'
                    )
        return retdict

    def get_composite_fields(self) -> dict[str, BaseFieldType]:
        return {field: self.owner.meta.fields[field] for field in self.inner_field_names}

    def is_required(self) -> bool:
        return False

    def has_default(self) -> bool:
        return False


class CompositeField(FieldFactory):
    """
    Meta field that aggregates multiple fields in a pseudo field
    """

    field_bases = (ConcreteCompositeField,)

    @classmethod
    def get_pydantic_type(cls, **kwargs: Any) -> Any:
        """Returns the type for pydantic"""
        if "model" in kwargs:
            return kwargs.get("model")
        return dict[str, Any]

    @classmethod
    def validate(cls, kwargs: dict[str, Any]) -> None:
        inner_fields = kwargs.get("inner_fields")
        if inner_fields is not None:
            if hasattr(inner_fields, "meta"):
                kwargs.setdefault("model", inner_fields)
                inner_fields = inner_fields.meta.fields
            if isinstance(inner_fields, dict):
                inner_fields = inner_fields.items()
            elif not isinstance(inner_fields, Sequence):
                raise FieldDefinitionError("inner_fields must be a Sequence, a dict or a model")
            if not inner_fields:
                raise FieldDefinitionError("inner_fields mustn't be empty")
            inner_field_names: set[str] = set()
            for field in inner_fields:
                if isinstance(field, str):
                    if field in inner_field_names:
                        raise FieldDefinitionError(f"duplicate inner field {field}")
                else:
                    if field[0] in inner_field_names:
                        raise FieldDefinitionError(f"duplicate inner field {field}")
        model = kwargs.get("model")
        if model is not None and not isinstance(model, type):
            raise FieldDefinitionError(f"model must be type {model}")



import contextlib
import copy
from abc import abstractmethod
from collections.abc import Sequence
from functools import cached_property
from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
    Optional,
    Union,
)

import sqlalchemy
from pydantic import BaseModel
from pydantic.fields import FieldInfo

from edgy.core.db.context_vars import CURRENT_PHASE, MODEL_GETATTR_BEHAVIOR
from edgy.types import Undefined

from .types import BaseFieldType, ColumnDefinitionModel

if TYPE_CHECKING:
    from edgy.core.connection.database import Database
    from edgy.core.connection.registry import Registry
    from edgy.core.db.models.types import BaseModelType


def _removesuffix(text: str, suffix: str) -> str:
    # TODO: replace with _removesuffix when python3.9 is minimum
    if text.endswith(suffix):
        return text[: -len(suffix)]
    else:
        return text


class BaseField(BaseFieldType, FieldInfo):
    """
    The base field for Edgy data model fields. It provides some helpers additional to
    BaseFieldType and inherits from FieldInfo for pydantic integration.

    Allows factories to overwrite methods.
    """

    # defs to simplify the life (can be None actually)
    owner: type["BaseModelType"]
    operator_mapping: dict[str, str] = {
        # aliases
        "is": "is_",
        "in": "in_",
        # this operators are not directly available and need their alias
        "exact": "__eq__",
        "not": "__ne__",
        "gt": "__gt__",
        "ge": "__ge__",
        "gte": "__ge__",
        "lt": "__lt__",
        "lte": "__le__",
        "le": "__le__",
    }

    def __init__(
        self,
        *,
        default: Any = Undefined,
        **kwargs: Any,
    ) -> None:
        if "__type__" in kwargs:
            kwargs["field_type"] = kwargs.pop("__type__")
        self.explicit_none = default is None

        super().__init__(**kwargs)

        # set remaining attributes
        for name, value in kwargs.items():
            setattr(self, name, value)

        # null is used for nullable columns and is_required is False
        # this is required for backward compatibility and pydantic_core uses null=True too
        # for opting out of nullable columns overwrite the get_column(s) method
        if (
            self.null or self.server_default is not None or self.autoincrement
        ) and default is Undefined:
            default = None
        if default is not Undefined:
            self.default = default

    def operator_to_clause(
        self, field_name: str, operator: str, table: sqlalchemy.Table, value: Any
    ) -> Any:
        """Base implementation, adaptable"""
        # Map the operation code onto SQLAlchemy's ColumnElement
        # https://docs.sqlalchemy.org/en/latest/core/sqlelement.html#sqlalchemy.sql.expression.ColumnElement
        # MUST raise an KeyError on missing columns, this code is used for the generic case if no field is available
        column = table.columns[field_name]
        operator = self.operator_mapping.get(operator, operator)
        if operator == "iexact":
            ESCAPE_CHARACTERS = ["%", "_"]
            has_escaped_character = any(c for c in ESCAPE_CHARACTERS if c in value)
            if has_escaped_character:
                value = value.replace("\\", "\\\\")
                # enable escape modifier
                for char in ESCAPE_CHARACTERS:
                    value = value.replace(char, f"\\{char}")
            clause = column.ilike(value, escape="\\" if has_escaped_character else None)
            return clause
        elif operator in {
            "contains",
            "icontains",
            "startswith",
            "endswith",
            "istartswith",
            "iendswith",
        }:
            return getattr(column, operator)(value, autoescape=True)
        return getattr(column, operator)(value)

    def is_required(self) -> bool:
        """Check if the argument is required.

        Returns:
            `True` if the argument is required, `False` otherwise.
        """
        if self.primary_key and self.autoincrement:
            return False
        return not (
            self.null
            or self.server_default is not None
            or ((self.default is not None or self.explicit_none) and self.default is not Undefined)
        )

    def has_default(self) -> bool:
        """Checks if the field has a default value set"""
        return bool(
            (self.default is not None or self.explicit_none) and self.default is not Undefined
        )

    def get_columns(self, name: str) -> Sequence[sqlalchemy.Column]:
        """
        Returns the columns of the field being declared.
        """
        return []

    def embed_field(
        self,
        prefix: str,
        new_fieldname: str,
        owner: Optional[type["BaseModelType"]] = None,
        parent: Optional["BaseField"] = None,
    ) -> Optional["BaseField"]:
        """
        Embed this field or return None to prevent embedding.
        Must return a copy with name and owner set when not returning None.
        """
        field_copy = copy.copy(self)
        field_copy.name = new_fieldname
        field_copy.owner = owner
        return field_copy

    def get_default_value(self) -> Any:
        # single default
        default = getattr(self, "default", None)
        if callable(default):
            return default()
        return default

    def get_default_values(self, field_name: str, cleaned_data: dict[str, Any]) -> dict[str, Any]:
        # for multidefaults overwrite in subclasses get_default_values to
        # parse default values differently
        # NOTE: multi value fields should always check here if defaults were already applied
        # NOTE: when build meta fields without columns this should be empty
        if field_name in cleaned_data:
            return {}
        return {field_name: self.get_default_value()}


class Field(BaseField):
    """
    Field with fallbacks and used for factories.
    """

    def check(self, value: Any) -> Any:
        """
        Runs the checks for the fields being validated. Single Column.
        """
        return value

    def clean(self, name: str, value: Any, for_query: bool = False) -> dict[str, Any]:
        """
        Runs the checks for the fields being validated. Multiple columns possible
        """
        return {name: self.check(value)}

    def get_column(self, name: str) -> Optional[sqlalchemy.Column]:
        """
        Return a single column for the field declared. Return None for meta fields.
        """
        column_model = ColumnDefinitionModel.model_validate(self, from_attributes=True)
        if column_model.column_type is None:
            return None
        return sqlalchemy.Column(
            column_model.column_name or name,
            column_model.column_type,
            *column_model.constraints,
            key=name,
            **column_model.model_dump(by_alias=True, exclude_none=True),
        )

    def get_columns(self, name: str) -> Sequence[sqlalchemy.Column]:
        column = self.get_column(name)
        if column is None:
            return []
        return [column]


class BaseCompositeField(BaseField):
    def translate_name(self, name: str) -> str:
        """translate name for inner objects and parsing values"""
        return name

    def get_composite_fields(self) -> dict[str, BaseFieldType]:
        """return dictionary of fields with untranslated names"""
        raise NotImplementedError()

    @cached_property
    def composite_fields(self) -> dict[str, BaseFieldType]:
        # return untranslated names
        return self.get_composite_fields()

    def clean(self, field_name: str, value: Any, for_query: bool = False) -> dict[str, Any]:
        """
        Runs the checks for the fields being validated.
        """
        prefix = _removesuffix(field_name, self.name)
        result = {}
        ErrorType: type[Exception] = KeyError
        if not isinstance(value, dict):
            # simpler
            value = value.__dict__
            # trigger load for missing attributes
            ErrorType = AttributeError

        for sub_name, field in self.composite_fields.items():
            translated_name = self.translate_name(sub_name)
            if translated_name not in value:
                if field.has_default() or not field.is_required():
                    continue
                raise ErrorType(f"Missing sub-field: {sub_name} for {field_name}")
            result.update(
                field.clean(f"{prefix}{sub_name}", value[translated_name], for_query=for_query)
            )
        return result

    def to_model(
        self,
        field_name: str,
        value: Any,
    ) -> dict[str, Any]:
        """
        Runs the checks for the fields being validated.
        """
        result = {}
        phase = CURRENT_PHASE.get()
        ErrorType: type[Exception] = KeyError
        if not isinstance(value, dict):
            # simpler
            value = value.__dict__
            # trigger load for missing attributes
            ErrorType = AttributeError
        for sub_name, field in self.composite_fields.items():
            translated_name = self.translate_name(sub_name)
            if translated_name not in value:
                if phase == "init" or phase == "init_db":
                    continue
                raise ErrorType(f"Missing sub-field: {sub_name} for {field_name}")
            result.update(field.to_model(sub_name, value.get(translated_name, None)))
        return result

    def get_default_values(self, field_name: str, cleaned_data: dict[str, Any]) -> Any:
        # fields should provide their own default which is used as long as they are in the fields mapping
        return {}


class RelationshipField(BaseField):
    def traverse_field(self, path: str) -> tuple[Any, str, str]:
        raise NotImplementedError()

    def is_cross_db(self, owner_database: Optional["Database"] = None) -> bool:
        raise NotImplementedError()


class PKField(BaseCompositeField):
    """
    Field for pk
    """

    def __init__(self, **kwargs: Any) -> None:
        kwargs["default"] = kwargs["server_default"] = None
        kwargs["field_type"] = kwargs["annotation"] = Any
        super().__init__(
            **kwargs,
        )

    def __get__(self, instance: "BaseModelType", owner: Any = None) -> Union[dict[str, Any], Any]:
        pkcolumns = self.owner.pkcolumns
        pknames = self.owner.pknames
        assert len(pkcolumns) >= 1
        # we don't want to issue loads
        token = MODEL_GETATTR_BEHAVIOR.set("passdown")
        try:
            if len(pknames) == 1:
                return getattr(instance, pknames[0], None)
            d = {}
            for key in pknames:
                translated_name = self.translate_name(key)
                field = instance.meta.fields.get(key)
                if field and hasattr(field, "__get__"):
                    d[translated_name] = field.__get__(instance, owner)
                else:
                    d[translated_name] = getattr(instance, key, None)
            for key in self.fieldless_pkcolumns:
                translated_name = self.translate_name(key)
                d[translated_name] = getattr(instance, key, None)
        finally:
            MODEL_GETATTR_BEHAVIOR.reset(token)
        return d

    def modify_input(self, name: str, kwargs: dict[str, Any]) -> None:
        if name not in kwargs:
            return
        # check for collisions
        for pkname in self.owner.pknames:
            if pkname in kwargs:
                raise ValueError("Cannot specify a primary key field and the primary key itself")

    def embed_field(
        self,
        prefix: str,
        new_fieldname: str,
        owner: Optional[type["BaseModelType"]] = None,
        parent: Optional[BaseFieldType] = None,
    ) -> Optional[BaseFieldType]:
        return None

    def clean(self, field_name: str, value: Any, for_query: bool = False) -> dict[str, Any]:
        pkcolumns = self.owner.pkcolumns
        pknames = self.owner.pknames
        prefix = _removesuffix(field_name, self.name)
        assert len(pkcolumns) >= 1
        if (
            len(pknames) == 1
            and not self.is_incomplete
            and not isinstance(value, (dict, BaseModel))
        ):
            pkname = pknames[0]
            field = self.owner.meta.fields[pkname]
            return field.clean(f"{prefix}{pkname}", value, for_query=for_query)
        retdict = super().clean(field_name, value, for_query=for_query)
        if self.is_incomplete:
            if isinstance(value, dict):
                for column_name in self.fieldless_pkcolumns:
                    translated_name = self.translate_name(column_name)
                    if translated_name not in value:
                        if not for_query:
                            continue
                        raise ValueError(f"Missing key: {translated_name} for {field_name}")
                    retdict[f"{prefix}{column_name}"] = value[translated_name]
            else:
                for column_name in self.fieldless_pkcolumns:
                    translated_name = self.translate_name(column_name)
                    if not hasattr(value, translated_name):
                        if not for_query:
                            continue
                        raise ValueError(f"Missing attribute: {translated_name} for {field_name}")
                    retdict[f"{prefix}{column_name}"] = getattr(value, translated_name)

        return retdict

    def to_model(
        self,
        field_name: str,
        value: Any,
    ) -> dict[str, Any]:
        pknames = self.owner.pknames
        assert len(self.owner.pkcolumns) >= 1
        if self.is_incomplete:
            raise ValueError("Cannot set an incomplete defined pk!")
        if (
            len(pknames) == 1
            # we first only redirect both
            and not isinstance(value, (dict, BaseModel))
        ):
            field = self.owner.meta.fields[pknames[0]]
            return field.to_model(pknames[0], value)
        return super().to_model(field_name, value)

    def get_composite_fields(self) -> dict[str, BaseFieldType]:
        return {field: self.owner.meta.fields[field] for field in self.owner.pknames}

    @cached_property
    def fieldless_pkcolumns(self) -> frozenset[str]:
        field_less = set()
        for colname in self.owner.pkcolumns:
            if colname not in self.owner.meta.columns_to_field:
                field_less.add(colname)
        return frozenset(field_less)

    @property
    def is_incomplete(self) -> bool:
        return bool(self.fieldless_pkcolumns)

    def is_required(self) -> bool:
        return False


class BaseForeignKey(RelationshipField):
    is_m2m: bool = False
    related_name: Union[str, Literal[False]] = ""
    # name used for backward relations
    # only useful if related_name = False because otherwise it gets overwritten
    reverse_name: str = ""

    @cached_property
    def target_registry(self) -> "Registry":
        """Registry searched in case to is a string"""
        assert self.owner.meta.registry, "no registry found neither 'target_registry' set"
        return self.owner.meta.registry

    @property
    def target(self) -> Any:
        """
        The target of the ForeignKey model.
        """
        if not hasattr(self, "_target"):
            if isinstance(self.to, str):
                self._target = self.target_registry.get_model(self.to)
            else:
                self._target = self.to
        return self._target

    @target.setter
    def target(self, value: Any) -> None:
        with contextlib.suppress(AttributeError):
            delattr(self, "_target")
        self.to = value

    @target.deleter
    def target(self, value: Any) -> None:
        with contextlib.suppress(AttributeError):
            delattr(self, "_target")

    def is_cross_db(self, owner_database: Optional["Database"] = None) -> bool:
        if owner_database is None:
            owner_database = self.owner.database
        return str(owner_database.url) != str(self.target.database.url)

    @abstractmethod
    def reverse_clean(self, name: str, value: Any, for_query: bool = False) -> dict[str, Any]: ...

    def expand_relationship(self, value: Any) -> Any:
        """
        Returns the related object or the relationship object
        """
        return value

    def to_model(
        self,
        field_name: str,
        value: Any,
    ) -> dict[str, Any]:
        return {field_name: self.expand_relationship(value)}



import os
import sys
from collections.abc import Generator, Sequence
from copy import copy
from functools import cached_property
from io import BytesIO
from typing import (
    TYPE_CHECKING,
    Any,
    BinaryIO,
    Callable,
    ClassVar,
    Literal,
    Optional,
    Union,
    cast,
)

from edgy.exceptions import FileOperationError, SuspiciousFileOperation

if TYPE_CHECKING:
    from .storage import Storage

if sys.version_info >= (3, 10):  # pragma: no cover
    from typing import ParamSpec
else:  # pragma: no cover
    from typing_extensions import ParamSpec

if TYPE_CHECKING:
    from PIL.ImageFile import ImageFile

    from edgy.core.db.fields.types import BaseFieldType
    from edgy.core.db.models.types import BaseModelType


P = ParamSpec("P")


def _get_storage(storage: str) -> "Storage":
    from .storage import storages

    return storages[storage]


class File:
    name: str
    file: Optional[BinaryIO]
    storage: "Storage"
    DEFAULT_CHUNK_SIZE: ClassVar[int] = 64 * 2**10
    mode: str = "rb"

    def __init__(
        self,
        file: Optional[BinaryIO] = None,
        name: str = "",
        storage: Union["Storage", str, None] = None,
    ) -> None:
        self.file = file
        if not storage:
            storage = "default"
        if isinstance(storage, str):
            storage = _get_storage(storage)

        self.storage = storage

        if not name:
            name = getattr(file, "name", "")

        self.name = name or ""

        if hasattr(file, "mode"):
            self.mode = file.mode

    def __eq__(self, other: Union[str, "File"]) -> bool:
        if hasattr(other, "name"):
            return self.name == other.name
        return self.name == other

    def __hash__(self) -> int:
        return hash(self._name)

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self or 'None'}>"

    def __bool__(self) -> bool:
        return bool(self.name or self.file is not None)

    @cached_property
    def size(self) -> int:
        if self.file is None:
            return 0
        if hasattr(self.file, "size"):
            return cast(int, self.file.size)
        if hasattr(self.file, "name"):
            try:
                return self.storage.size(self.file.name)
            except (OSError, TypeError, SuspiciousFileOperation):
                # we can have a reference outside of media folder so ignore SuspiciousFileOperation
                pass
        if hasattr(self.file, "tell") and hasattr(self.file, "seek"):
            pos = self.file.tell()
            self.file.seek(0, os.SEEK_END)
            size = self.file.tell()
            self.file.seek(pos)
            return size
        raise AttributeError("Unable to determine the file's size.")

    def __len__(self) -> int:
        return self.size

    @property
    def closed(self) -> bool:
        """Return True if the file is closed."""
        return not self.file or self.file.closed

    @property
    def path(self) -> str:
        return self.storage.path(self.name)

    @property
    def url(self) -> str:
        return self.storage.url(self.name)

    def chunks(self, chunk_size: Union[int, None] = None) -> Generator[bytes, None, None]:
        """
        Read the file and yield chunks of ``chunk_size`` bytes (defaults to
        ``File.DEFAULT_CHUNK_SIZE``).

        Args:
            chunk_size (int, optional): The size of each chunk in bytes. Defaults to None.

        Yields:
            bytes: Data read from the file.

        Raises:
            AttributeError: If unable to seek to the beginning of the file.
        """
        assert self.file is not None, "File is closed"
        chunk_size = chunk_size or self.DEFAULT_CHUNK_SIZE

        try:
            self.file.seek(0)
        except (AttributeError, OSError):
            ...
        else:
            while True:
                data = self.file.read(chunk_size)
                if not data:
                    break
                yield data

    def multiple_chunks(self, chunk_size: Union[int, None] = None) -> bool:
        """
        Return ``True`` if you can expect multiple chunks.

        Args:
            chunk_size (int, optional): The size of each chunk in bytes. Defaults to None.

        Returns:
            bool: True if the file can be expected to have multiple chunks, False otherwise.
        """
        if chunk_size is None:
            chunk_size = self.DEFAULT_CHUNK_SIZE

        return self.size > chunk_size

    def __enter__(self) -> "File":
        assert self.file is not None, "File is closed"
        return self

    def __exit__(self, exc_type: Exception, exc_value: Any, tb: Any) -> None:
        self.close()

    def open(self, mode: Union[str, None] = None) -> "File":
        """
        Open the file with the specified mode.

        Args:
            mode (str, optional): The mode in which to open the file. If not provided, uses the existing mode.

        Returns:
            File: The opened file.

        Raises:
            FileOperationError: If the file cannot be reopened.
        """
        if not self.closed:
            self.file.seek(0)
        elif self.name and self.storage.exists(self.name):
            self.file = self.storage.open(self.name, mode or self.mode)  # noqa: SIM115
        else:
            raise FileOperationError("The file cannot be reopened.")

        return self

    def readable(self) -> bool:
        """Return True if the file is readable."""
        if self.closed:
            return False
        assert self.file is not None
        if hasattr(self.file, "readable"):
            return self.file.readable()
        return True

    def writable(self) -> bool:
        """Return True if the file is writable."""
        if self.closed:
            return False
        assert self.file is not None
        if hasattr(self.file, "writable"):
            return self.file.writable()
        return "w" in getattr(self.file, "mode", "")

    def seekable(self) -> bool:
        """Return True if the file is seekable."""
        if self.closed:
            return False
        assert self.file is not None
        if hasattr(self.file, "seekable"):
            return self.file.seekable()
        return True

    def seek(self, offset: int, whence: int = 0) -> int:
        assert self.seekable()
        return self.file.seek(offset, whence)

    def tell(self) -> int:
        assert self.file is not None
        return self.file.tell()

    def read(self, amount: Optional[int] = None) -> bytes:
        assert self.file is not None
        return self.file.read(amount)

    def write(self, data: bytes) -> int:
        assert self.file is not None
        self.__dict__.pop("size", None)
        return self.file.write(data)

    def close(self, keep_size: bool = False) -> None:
        if self.file is None:
            return
        self.file.close()
        self.file = None
        if not keep_size:
            self.__dict__.pop("size", None)


class ContentFile(File):
    file: BinaryIO

    def __init__(self, content: bytes, name: str = ""):
        super().__init__(file=BytesIO(content), name=name)
        self.size = len(content)

    def __str__(self) -> str:
        return "Raw content"

    def open(self, mode: Union[str, Any] = None) -> "ContentFile":
        self.file.seek(0)
        return self

    def close(self, keep_size: bool = False) -> None:
        if not keep_size:
            self.__dict__.pop("size", None)


class FieldFile(File):
    operation: Literal["none", "save", "save_delete", "delete"] = "none"
    old: Optional[tuple["Storage", str, bool]] = None
    instance: Optional["BaseModelType"] = None
    # can extract metadata
    approved: bool
    metadata: dict[str, Any]

    def __init__(
        self,
        field: "BaseFieldType",
        content: Union[BinaryIO, bytes, None, File] = None,
        name: str = "",
        size: Optional[int] = None,
        storage: Union["Storage", str, None] = None,
        generate_name_fn: Optional[Callable[[str, Union[BinaryIO, File], bool], str]] = None,
        metadata: Optional[dict[str, Any]] = None,
        multi_process_safe: bool = True,
        approved: bool = True,
        # only usable with correct approval handling
        change_removes_approval: bool = False,
    ) -> None:
        if isinstance(content, File):
            content = content.open("rb").file
        elif isinstance(content, bytes):
            content = BytesIO(content)
        super().__init__(content, name=name, storage=storage)
        self.field = field
        self.generate_name_fn = generate_name_fn
        self.metadata = metadata or {}
        self.multi_process_safe = multi_process_safe
        self.change_removes_approval = change_removes_approval
        self.approved = approved
        if size is not None:
            # set value to cached_property
            self.size = size

    def to_file(self) -> Optional[File]:
        """Cloak FileField so it looks like a regular File. Required for copies."""
        if self:
            return File(cast(BinaryIO, copy(self)), name=self.name, storage=self.storage)
        return None

    async def execute_operation(self, nodelete_old: bool = False) -> None:
        operation = self.operation
        self.operation = "none"
        if operation == "save" or operation == "save_delete":
            if self.file is None or not self.name:
                raise ValueError(
                    f"The '{self.field.name}' attribute has no file associated with it."
                )
            try:
                self.storage.save(self.file, self.name)
            finally:
                self.storage.unreserve_name(self.name)
            if (
                not nodelete_old
                and operation == "save_delete"
                and self.old is not None
                and self.old[1]
                and self.old[1] != self.name
            ):
                self.old[0].delete(self.old[1])
        elif operation == "delete":
            if getattr(self, "file", None):
                self.close()
            # old should not be None anyway but check that
            # if name is empty or None skip deletion
            if not nodelete_old and self.old is not None and self.old[1]:
                self.old[0].delete(self.old[1])
        # else approve operation or metadata update
        self.old = None

    def save(
        self,
        content: Union[BinaryIO, bytes, None, File],
        *,
        name: str = "",
        delete_old: bool = True,
        multi_process_safe: Optional[bool] = None,
        approved: Optional[bool] = None,
        storage: Optional["Storage"] = None,
        overwrite: bool = False,
    ) -> None:
        """
        Save the file to storage and update associated model fields.

        Args:
            name (str): The name of the file.
            content: The file content.
        """
        if content is None:
            self.delete()
            return
        # we can force multi_process_safe to add the process id also for overwrites
        if multi_process_safe is None:
            multi_process_safe = False if overwrite else self.multi_process_safe

        direct_name = True
        if not name:
            direct_name = False
            name = getattr(content, "name", "")

        if isinstance(content, File):
            content = content.open("rb").file
        elif isinstance(content, bytes):
            content = BytesIO(content)

        # Generate filename based on name
        if self.generate_name_fn is not None:
            name = self.generate_name_fn(name, content, direct_name)

        assert name, "no name found"

        if storage is None:
            storage = self.storage

        # filters invalid names
        name = storage.get_available_name(
            name,
            max_length=getattr(self.field, "max_length", None),
            overwrite=overwrite,
            multi_process_safe=multi_process_safe,
        )
        if getattr(self, "file", None):
            self.close()
        self.file = content
        self.old = (self.storage, self.name, self.approved)
        self.storage = storage
        self.name = name
        self.operation = "save_delete" if delete_old else "save"
        if approved is not None:
            self.approved = approved
        elif self.change_removes_approval:
            self.approved = False

    def set_approved(self, approved: bool) -> None:
        self.old = (self.storage, self.name, self.approved)
        self.approved = approved

    def reset(self) -> None:
        """
        Reset staged file operation.
        """
        if self.operation == "save" or self.operation == "save_delete":
            # delete new assigned file
            if getattr(self, "file", None):
                self.close()
            self.storage.unreserve_name(self.name)
        if self.old is not None:
            self.storage, self.name, self.approved = self.old
            self.old = None
        self.operation = "none"

    def delete(self, *, approved: Optional[bool] = None, instant: bool = False) -> None:
        """
        Mark the file associated with this object for deletion from storage.
        """
        if not self:
            return
        if not self.field.null and not instant:
            raise FileOperationError("Cannot delete file (only replacing is possible)")
        self.reset()
        if instant:
            # close
            if getattr(self, "file", None):
                self.close()
            self.storage.delete(self.name)
            # when allowed setting null, do so
            # this function must be callable by model deletions
            if self.field.null:
                self.name = ""
            if approved is not None:
                self.approved = approved
            elif self.change_removes_approval:
                self.approved = False
            return
        self.old = (self.storage, self.name, self.approved)
        self.name = ""
        self.operation = "delete"
        if approved is not None:
            self.approved = approved
        elif self.change_removes_approval:
            self.approved = False

    def close(self, keep_size: bool = False) -> None:
        if self.operation == "none":
            super().close(keep_size=keep_size)


class ImageFieldFile(FieldFile):
    def open_image(self) -> "ImageFile":
        from PIL import Image

        allowed_formats: Optional[Sequence[str]] = getattr(self.field, "image_formats", ())
        if self.approved and allowed_formats is not None:
            approved_image_formats: Optional[Sequence[str]] = getattr(
                self.field, "approved_image_formats", ()
            )
            if approved_image_formats is None:
                allowed_formats = None
            else:
                allowed_formats = (*allowed_formats, *approved_image_formats)
        return Image.open(self.open("rb"), formats=allowed_formats)  # type: ignore



import os
import pathlib
import re
import secrets
import string
from os.path import abspath, dirname, join, normcase, sep
from typing import Any, Optional
from urllib.parse import quote

from edgy.exceptions import SuspiciousFileOperation


def safe_join(base: str, *paths: Any) -> str:
    final_path = _get_final_path(base, *paths)
    base_path = abspath(base)
    _validate_final_path(final_path, base_path)
    return final_path


def _get_final_path(base: str, *paths: Any) -> str:
    """
    Get the final path by joining base and other paths.

    Args:
        base (str): The base path.
        paths (str): Additional paths to join with the base path.

    Returns:
        str: The final joined path.
    """
    return abspath(join(base, *paths))


def _validate_final_path(final_path: str, base_path: str) -> None:
    """
    Validate the final path against the base path.

    Args:
        final_path (str): The final joined path.
        base_path (str): The base path.

    Raises:
        SuspiciousFileOperation: If the final path is not located inside the base path component.
    """
    if (
        not normcase(final_path).startswith(normcase(base_path + sep))
        and normcase(final_path) != normcase(base_path)
        and dirname(normcase(base_path)) != normcase(base_path)
    ):
        raise SuspiciousFileOperation(
            f"The joined path ({final_path}) is located outside of the base path "
            f"component ({base_path})"
        )


def get_valid_filename(name: str) -> str:
    """
    Cleans and validates a string to be used as a filename.

    Args:
        name (str): The string to be cleaned.

    Raises:
        SuspiciousFileOperation: If the resulting filename is empty or consists of '.', or '..'.

    Returns:
        str: The cleaned and validated filename.
    """
    # Strip leading and trailing whitespaces, and replace inner whitespaces with underscores
    cleaned_name = str(name).strip().replace(" ", "_")

    # Remove any characters that are not letters, numbers, hyphens, underscores, or dots
    cleaned_name = re.sub(r"(?u)[^-\w.]", "", cleaned_name)

    # Check if the resulting filename is empty or consists of '.', or '..'
    if cleaned_name in {"", ".", ".."}:
        raise SuspiciousFileOperation(f"Could not derive file name from '{name}'")

    return cleaned_name


def get_random_string(length: int = 10) -> str:
    """
    Generates a random string of specified length.

    Args:
        length (int, optional): The length of the random string. Defaults to 10.

    Returns:
        str: The generated random string.
    """
    letters = string.ascii_lowercase
    random_string = "".join(secrets.choice(letters) for _ in range(length))
    return random_string


def validate_file_name(name: str, allow_relative_path: bool = False) -> str:
    """
    Validate a file name to ensure it is safe and does not contain path traversal attempts.

    Args:
        name (str): The file name to be validated.
        allow_relative_path (bool, optional): Whether to allow relative paths. Defaults to False.

    Raises:
        SuspiciousFileOperation: If the file name is potentially dangerous or contains path traversal attempts.

    Returns:
        str: The validated file name.
    """
    base_name = os.path.basename(name)
    if base_name in {"", ".", ".."}:
        raise SuspiciousFileOperation(f"Could not derive file name from '{name}'")

    if allow_relative_path:
        path = pathlib.PurePosixPath(name)
        if path.is_absolute() or ".." in path.parts:
            raise SuspiciousFileOperation(f"Detected path traversal attempt in '{name}'")
    elif name != base_name:
        raise SuspiciousFileOperation(f"File name '{name}' includes path elements")

    return name


def filepath_to_uri(path: Optional[str]) -> str:
    """
    Convert a file system path to a URI portion suitable for inclusion in a URL.

    Encode certain characters that would normally be recognized as special characters
    for URIs. Do not encode the ' character, as it is a valid character
    within URIs. See the encodeURIComponent() JavaScript function for details.

    Args:
        path (str): The file system path to convert.

    Returns:
        str: The URI portion suitable for inclusion in a URL.
    """
    if not path:
        return ""
    return quote(str(path).replace("\\", "/"), safe="/~!*()'")



import datetime
import decimal
import enum
import ipaddress
import uuid
import warnings
from collections.abc import Callable, Sequence
from enum import EnumMeta
from functools import cached_property, partial
from re import Pattern
from typing import TYPE_CHECKING, Any, Optional, Union, cast

import pydantic
import sqlalchemy
from pydantic import EmailStr
from sqlalchemy.dialects import oracle

from edgy.core.db.context_vars import CURRENT_INSTANCE, CURRENT_PHASE, EXPLICIT_SPECIFIED_VALUES
from edgy.core.db.fields._internal import IPAddress
from edgy.core.db.fields._validators import IPV4_REGEX, IPV6_REGEX
from edgy.core.db.fields.base import BaseField, Field
from edgy.core.db.fields.factories import FieldFactory
from edgy.core.db.fields.types import BaseFieldType
from edgy.exceptions import FieldDefinitionError

if TYPE_CHECKING:
    import zoneinfo

    from edgy.core.db.models.types import BaseModelType


CLASS_DEFAULTS = ["cls", "__class__", "kwargs"]


class ComputedField(BaseField):
    def __init__(
        self,
        getter: Union[
            Callable[[BaseFieldType, "BaseModelType", Optional[type["BaseModelType"]]], Any], str
        ],
        setter: Union[Callable[[BaseFieldType, "BaseModelType", Any], None], str, None] = None,
        fallback_getter: Optional[
            Callable[[BaseFieldType, "BaseModelType", Optional[type["BaseModelType"]]], Any]
        ] = None,
        **kwargs: Any,
    ) -> None:
        kwargs["exclude"] = True
        kwargs["null"] = True
        kwargs["primary_key"] = False
        kwargs["field_type"] = kwargs["annotation"] = Any
        self.getter = getter
        self.fallback_getter = fallback_getter
        self.setter = setter
        super().__init__(
            **kwargs,
        )

    @cached_property
    def compute_getter(
        self,
    ) -> Callable[[BaseFieldType, "BaseModelType", Optional[type["BaseModelType"]]], Any]:
        if isinstance(self.getter, str):
            fn = cast(
                Optional[
                    Callable[
                        [BaseFieldType, "BaseModelType", Optional[type["BaseModelType"]]], Any
                    ]
                ],
                getattr(self.owner, self.getter, None),
            )
        else:
            fn = self.getter
        if fn is None and self.fallback_getter is not None:
            fn = self.fallback_getter
        if fn is None:
            raise ValueError(f"No getter found for attribute: {self.getter}.")
        return fn

    @cached_property
    def compute_setter(self) -> Callable[[BaseFieldType, "BaseModelType", Any], None]:
        if isinstance(self.setter, str):
            fn = cast(
                Optional[Callable[[BaseFieldType, "BaseModelType", Any], None]],
                getattr(self.owner, self.setter, None),
            )
        else:
            fn = self.setter
        if fn is None:
            return lambda instance, name, value: None
        return fn

    def to_model(
        self,
        field_name: str,
        value: Any,
    ) -> dict[str, Any]:
        return {}

    def clean(
        self,
        name: str,
        value: Any,
        for_query: bool = False,
    ) -> dict[str, Any]:
        return {}

    def __get__(self, instance: "BaseModelType", owner: Any = None) -> Any:
        return self.compute_getter(self, instance, owner)

    def __set__(self, instance: "BaseModelType", value: Any) -> None:
        self.compute_setter(self, instance, value)


class CharField(FieldFactory, str):
    """String field representation that constructs the Field class and populates the values"""

    field_type = str

    def __new__(  # type: ignore
        cls,
        *,
        min_length: Optional[int] = None,
        regex: Union[str, Pattern] = None,
        pattern: Union[str, Pattern] = None,
        **kwargs: Any,
    ) -> BaseFieldType:
        if pattern is None:
            pattern = regex
        del regex
        kwargs = {
            **kwargs,
            **{key: value for key, value in locals().items() if key not in CLASS_DEFAULTS},
        }

        return super().__new__(cls, **kwargs)

    @classmethod
    def validate(cls, kwargs: dict[str, Any]) -> None:
        max_length = kwargs.get("max_length", 0)
        if max_length is not None and max_length <= 0:
            raise FieldDefinitionError(detail=f"'max_length' is required for {cls.__name__}")

        min_length = kwargs.get("min_length")
        pattern = kwargs.get("regex")

        assert min_length is None or isinstance(min_length, int)
        assert max_length is None or isinstance(max_length, int)
        assert pattern is None or isinstance(pattern, (str, Pattern))

    @classmethod
    def get_column_type(cls, max_length: Optional[int] = None, **kwargs: Any) -> Any:
        return (
            sqlalchemy.Text(collation=kwargs.get("collation"))
            if max_length is None
            else sqlalchemy.String(length=max_length, collation=kwargs.get("collation"))
        )


class TextField(CharField):
    """String representation of a text field which means no max_length required"""

    @classmethod
    def validate(cls, kwargs: dict[str, Any]) -> None:
        kwargs.setdefault("max_length", None)
        super().validate(kwargs)


class IncrementOnSaveBaseField(Field):
    increment_on_save: int = 0

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(
            **kwargs,
        )
        if self.increment_on_save != 0:
            self.pre_save_callback = self._notset_pre_save_callback

    async def _notset_pre_save_callback(
        self, value: Any, original_value: Any, force_insert: bool, instance: "BaseModelType"
    ) -> dict[str, Any]:
        explicit_values = EXPLICIT_SPECIFIED_VALUES.get()
        if explicit_values is not None and self.name in explicit_values:
            return {}
        model_or_query = CURRENT_INSTANCE.get()

        if force_insert:
            if original_value is None:
                return {self.name: self.get_default_value()}
            else:
                return {self.name: value + self.increment_on_save}
        elif not self.primary_key:
            # update path
            return {
                self.name: (
                    model_or_query if model_or_query is not None else instance
                ).table.columns[self.name]
                + self.increment_on_save
            }
        else:
            # update path
            return {}

    def get_default_values(
        self,
        field_name: str,
        cleaned_data: dict[str, Any],
    ) -> dict[str, Any]:
        if self.increment_on_save != 0:
            phase = CURRENT_PHASE.get()
            if phase in "prepare_update":
                return {field_name: None}
        return super().get_default_values(field_name, cleaned_data)

    def to_model(
        self,
        field_name: str,
        value: Any,
    ) -> dict[str, Any]:
        phase = CURRENT_PHASE.get()
        instance = CURRENT_INSTANCE.get()
        if self.increment_on_save != 0 and not self.primary_key and phase == "post_update":
            # a bit dirty but works
            instance.__dict__.pop(field_name, None)
            return {}
        return super().to_model(field_name, value)


class IntegerField(FieldFactory, int):
    """
    Integer field factory that construct Field classes and populated their values.
    """

    field_type = int
    field_bases = (IncrementOnSaveBaseField,)

    def __new__(  # type: ignore
        cls,
        *,
        ge: Union[int, float, decimal.Decimal, None] = None,
        gt: Union[int, float, decimal.Decimal, None] = None,
        le: Union[int, float, decimal.Decimal, None] = None,
        lt: Union[int, float, decimal.Decimal, None] = None,
        multiple_of: Optional[int] = None,
        increment_on_save: int = 0,
        **kwargs: Any,
    ) -> BaseFieldType:
        kwargs = {
            **kwargs,
            **{k: v for k, v in locals().items() if k not in ["cls", "__class__", "kwargs"]},
        }
        return super().__new__(cls, **kwargs)

    @classmethod
    def get_column_type(cls, **kwargs: Any) -> Any:
        return sqlalchemy.Integer()

    @classmethod
    def validate(cls, kwargs: dict[str, Any]) -> None:
        increment_on_save = kwargs.get("increment_on_save", 0)
        if (
            increment_on_save == 0
            and kwargs.get("primary_key", False)
            and "autoincrement" not in kwargs
        ):
            warnings.warn(
                (
                    "Not setting autoincrement on an Integer field with `primary_key=True` is deprecated."
                    "We change the default to False in future."
                ),
                DeprecationWarning,
                stacklevel=4,
            )
            kwargs["autoincrement"] = True
        if increment_on_save != 0:
            if kwargs.get("autoincrement"):
                raise FieldDefinitionError(
                    detail="'autoincrement' is incompatible with 'increment_on_save'"
                )
            if kwargs.get("null"):
                raise FieldDefinitionError(
                    detail="'null' is incompatible with 'increment_on_save'"
                )
            kwargs.setdefault("read_only", True)
            kwargs["inject_default_on_partial_update"] = True


class FloatField(FieldFactory, float):
    """Representation of a int32 and int64"""

    field_type = float

    def __new__(  # type: ignore
        cls,
        *,
        max_digits: Optional[int] = None,
        ge: Union[int, float, decimal.Decimal, None] = None,
        gt: Union[int, float, decimal.Decimal, None] = None,
        le: Union[int, float, decimal.Decimal, None] = None,
        lt: Union[int, float, decimal.Decimal, None] = None,
        **kwargs: Any,
    ) -> BaseFieldType:
        # pydantic doesn't support max_digits for float, so rename it
        if max_digits is not None:
            kwargs.setdefault("precision", max_digits)
        del max_digits
        kwargs = {
            **kwargs,
            **{key: value for key, value in locals().items() if key not in CLASS_DEFAULTS},
        }
        return super().__new__(cls, **kwargs)

    @classmethod
    def get_column_type(cls, precision: Optional[int] = None, **kwargs: Any) -> Any:
        if precision is None:
            return sqlalchemy.Float(asdecimal=False)
        return sqlalchemy.Float(precision=precision, asdecimal=False).with_variant(
            oracle.FLOAT(binary_precision=round(precision / 0.30103), asdecimal=False),  # type: ignore
            "oracle",
        )


class BigIntegerField(IntegerField):
    """Representation of big integer field"""

    @classmethod
    def validate(cls, kwargs: dict[str, Any]) -> None:
        super().validate(kwargs)
        if kwargs.get("autoincrement", False):
            kwargs.setdefault("skip_reflection_type_check", True)

    @classmethod
    def get_column_type(cls, **kwargs: Any) -> Any:
        # sqlite special we cannot have a big IntegerField as PK
        if kwargs.get("autoincrement"):
            return sqlalchemy.BigInteger().with_variant(sqlalchemy.Integer, "sqlite")
        return sqlalchemy.BigInteger()


class SmallIntegerField(IntegerField):
    """Represents a small integer field"""

    @classmethod
    def get_column_type(cls, **kwargs: Any) -> Any:
        return sqlalchemy.SmallInteger()


class DecimalField(FieldFactory, decimal.Decimal):
    field_type = decimal.Decimal

    def __new__(  # type: ignore
        cls,
        *,
        ge: Union[int, float, decimal.Decimal, None] = None,
        gt: Union[int, float, decimal.Decimal, None] = None,
        le: Union[int, float, decimal.Decimal, None] = None,
        lt: Union[int, float, decimal.Decimal, None] = None,
        multiple_of: Union[int, decimal.Decimal, None] = None,
        max_digits: Optional[int] = None,
        decimal_places: Optional[int] = None,
        **kwargs: Any,
    ) -> BaseFieldType:
        kwargs = {
            **kwargs,
            **{k: v for k, v in locals().items() if k not in ["cls", "__class__", "kwargs"]},
        }
        return super().__new__(cls, **kwargs)

    @classmethod
    def get_column_type(cls, **kwargs: Any) -> Any:
        return sqlalchemy.Numeric(
            precision=kwargs.get("max_digits"), scale=kwargs.get("decimal_places"), asdecimal=True
        )

    @classmethod
    def validate(cls, kwargs: dict[str, Any]) -> None:
        super().validate(kwargs)

        decimal_places = kwargs.get("decimal_places")
        if decimal_places is None or decimal_places < 0:
            raise FieldDefinitionError("decimal_places are required for DecimalField")


class BooleanField(FieldFactory, int):
    """Representation of a boolean"""

    field_type = bool

    def __new__(  # type: ignore
        cls,
        *,
        default: Optional[bool] = False,
        **kwargs: Any,
    ) -> BaseFieldType:
        kwargs = {
            **kwargs,
            **{key: value for key, value in locals().items() if key not in CLASS_DEFAULTS},
        }
        return super().__new__(cls, **kwargs)

    @classmethod
    def get_column_type(cls, **kwargs: Any) -> Any:
        return sqlalchemy.Boolean()


class TimezonedField:
    default_timezone: Optional["zoneinfo.ZoneInfo"]
    force_timezone: Optional["zoneinfo.ZoneInfo"]
    remove_timezone: bool

    def _convert_datetime(
        self, value: datetime.datetime
    ) -> Union[datetime.datetime, datetime.date]:
        if value.tzinfo is None and self.default_timezone is not None:
            value = value.replace(tzinfo=self.default_timezone)
        if self.force_timezone is not None and value.tzinfo != self.force_timezone:
            if value.tzinfo is None:
                value = value.replace(tzinfo=self.force_timezone)
            else:
                value = value.astimezone(self.force_timezone)
        if self.remove_timezone:
            value = value.replace(tzinfo=None)
        if self.field_type is datetime.date:
            return value.date()
        return value

    def check(self, value: Any) -> Optional[Union[datetime.datetime, datetime.date]]:
        if value is None:
            return None
        elif isinstance(value, datetime.datetime):
            return self._convert_datetime(value)
        elif isinstance(value, (int, float)):
            return self._convert_datetime(
                datetime.datetime.fromtimestamp(value, self.default_timezone)
            )
        elif isinstance(value, str):
            return self._convert_datetime(datetime.datetime.fromisoformat(value))
        elif isinstance(value, datetime.date):
            # datetime is subclass, so check datetime first

            # don't touch dates when DateField
            if self.field_type is datetime.date:
                return value
            return self._convert_datetime(
                datetime.datetime(year=value.year, month=value.month, day=value.day)
            )
        else:
            raise ValueError(f"Invalid type detected: {type(value)}")

    def to_model(
        self,
        field_name: str,
        value: Any,
    ) -> dict[str, Optional[Union[datetime.datetime, datetime.date]]]:
        """
        Convert input object to datetime
        """
        return {field_name: self.check(value)}

    def get_default_value(self) -> Any:
        return self.check(super().get_default_value())


class AutoNowMixin(FieldFactory):
    def __new__(  # type: ignore
        cls,
        *,
        auto_now: Optional[bool] = False,
        auto_now_add: Optional[bool] = False,
        default_timezone: Optional["zoneinfo.ZoneInfo"] = None,
        **kwargs: Any,
    ) -> BaseFieldType:
        if auto_now_add and auto_now:
            raise FieldDefinitionError("'auto_now' and 'auto_now_add' cannot be both True")

        if auto_now_add or auto_now:
            kwargs.setdefault("read_only", True)
            kwargs["inject_default_on_partial_update"] = auto_now

        kwargs = {
            **kwargs,
            **{k: v for k, v in locals().items() if k not in CLASS_DEFAULTS},
        }
        if auto_now_add or auto_now:
            # date.today cannot handle timezone so use alway datetime and convert back to date
            kwargs["default"] = partial(datetime.datetime.now, default_timezone)
        return super().__new__(cls, **kwargs)


class DateTimeField(AutoNowMixin, datetime.datetime):
    """Representation of a datetime field"""

    field_type = datetime.datetime
    field_bases = (TimezonedField, Field)

    def __new__(  # type: ignore
        cls,
        *,
        auto_now: Optional[bool] = False,
        auto_now_add: Optional[bool] = False,
        default_timezone: Optional["zoneinfo.ZoneInfo"] = None,
        force_timezone: Optional["zoneinfo.ZoneInfo"] = None,
        remove_timezone: bool = False,
        **kwargs: Any,
    ) -> BaseFieldType:
        kwargs = {
            **kwargs,
            **{k: v for k, v in locals().items() if k not in CLASS_DEFAULTS},
        }
        kwargs.setdefault("with_timezone", not remove_timezone)
        return super().__new__(cls, **kwargs)

    @classmethod
    def get_column_type(cls, with_timezone: bool = True, **kwargs: Any) -> Any:
        return sqlalchemy.DateTime(with_timezone)

    @classmethod
    def get_default_values(
        cls,
        field_obj: Field,
        field_name: str,
        cleaned_data: dict[str, Any],
        original_fn: Any = None,
    ) -> Any:
        phase = CURRENT_PHASE.get()
        if field_obj.auto_now_add and phase == "prepare_update":
            return {}
        return original_fn(field_name, cleaned_data)


class DateField(AutoNowMixin, datetime.date):
    """Representation of a date field"""

    field_type = datetime.date
    field_bases = (TimezonedField, Field)

    def __new__(  # type: ignore
        cls,
        *,
        auto_now: Optional[bool] = False,
        auto_now_add: Optional[bool] = False,
        default_timezone: Optional["zoneinfo.ZoneInfo"] = None,
        force_timezone: Optional["zoneinfo.ZoneInfo"] = None,
        **kwargs: Any,
    ) -> BaseFieldType:
        # the datetimes lose the information anyway
        kwargs["remove_timezone"] = False

        kwargs = {
            **kwargs,
            **{k: v for k, v in locals().items() if k not in CLASS_DEFAULTS},
        }
        return super().__new__(cls, **kwargs)

    @classmethod
    def get_column_type(cls, **kwargs: Any) -> Any:
        return sqlalchemy.Date()


class DurationField(FieldFactory, datetime.timedelta):
    """Representation of a time field"""

    field_type = datetime.timedelta

    @classmethod
    def get_column_type(cls, **kwargs: Any) -> Any:
        return sqlalchemy.Interval()


class TimeField(FieldFactory, datetime.time):
    """Representation of a time field"""

    field_type = datetime.time

    def __new__(cls, with_timezone: bool = False, **kwargs: Any) -> BaseFieldType:  # type: ignore
        kwargs = {
            **kwargs,
            **{k: v for k, v in locals().items() if k not in CLASS_DEFAULTS},
        }
        return super().__new__(cls, **kwargs)

    @classmethod
    def get_column_type(cls, **kwargs: Any) -> Any:
        return sqlalchemy.Time(kwargs.get("with_timezone") or False)


class JSONField(FieldFactory, pydantic.Json):  # type: ignore
    """Representation of a JSONField"""

    field_type = Any

    @classmethod
    def get_column_type(cls, **kwargs: Any) -> Any:
        return sqlalchemy.JSON()


class BinaryField(FieldFactory, bytes):
    """Representation of a binary"""

    field_type = bytes

    def __new__(cls, *, max_length: Optional[int] = None, **kwargs: Any) -> BaseFieldType:  # type: ignore
        kwargs = {
            **kwargs,
            **{k: v for k, v in locals().items() if k not in CLASS_DEFAULTS},
        }
        return super().__new__(cls, **kwargs)

    @classmethod
    def get_column_type(cls, **kwargs: Any) -> Any:
        return sqlalchemy.LargeBinary(length=kwargs.get("max_length"))


class UUIDField(FieldFactory, uuid.UUID):
    """Representation of a uuid"""

    field_type = uuid.UUID

    def __new__(cls, **kwargs: Any) -> BaseFieldType:  # type: ignore
        kwargs = {
            **kwargs,
            **{k: v for k, v in locals().items() if k not in CLASS_DEFAULTS},
        }

        return super().__new__(cls, **kwargs)

    @classmethod
    def get_column_type(cls, **kwargs: Any) -> Any:
        return sqlalchemy.Uuid(as_uuid=True, native_uuid=True)


class ChoiceField(FieldFactory):
    """Representation of an Enum"""

    field_type = enum.Enum

    def __new__(  # type: ignore
        cls,
        choices: Optional[Sequence[Union[tuple[str, str], tuple[str, int]]]] = None,
        **kwargs: Any,
    ) -> BaseFieldType:
        kwargs = {
            **kwargs,
            **{k: v for k, v in locals().items() if k not in CLASS_DEFAULTS},
        }
        return super().__new__(cls, **kwargs)

    @classmethod
    def validate(cls, kwargs: dict[str, Any]) -> None:
        choice_class = kwargs.get("choices")
        if choice_class is None or not isinstance(choice_class, EnumMeta):
            raise FieldDefinitionError("ChoiceField choices must be an Enum")

    @classmethod
    def get_column_type(cls, **kwargs: Any) -> sqlalchemy.Enum:
        choice_class = kwargs.get("choices")
        return sqlalchemy.Enum(choice_class)


class PasswordField(CharField):
    """
    Representation of a Password
    """

    def __new__(  # type: ignore
        cls,
        derive_fn: Optional[Callable[[str], str]] = None,
        **kwargs: Any,
    ) -> BaseFieldType:
        kwargs.setdefault("keep_original", derive_fn is not None)
        return super().__new__(cls, derive_fn=derive_fn, **kwargs)

    @classmethod
    def to_model(
        cls, field_obj: BaseFieldType, field_name: str, value: Any, original_fn: Any = None
    ) -> dict[str, Any]:
        if isinstance(value, (tuple, list)):
            if value[0] != value[1]:
                raise ValueError("Password doesn't match.")
            else:
                value = value[0]
        retval: dict[str, Any] = {}
        phase = CURRENT_PHASE.get()
        derive_fn = cast(Optional[Callable[[str], str]], field_obj.derive_fn)
        if phase in {"set", "init"} and derive_fn is not None:
            retval[field_name] = derive_fn(value)
            if getattr(field_obj, "keep_original", False):
                retval[f"{field_name}_original"] = value
        else:
            retval[field_name] = value
            # blank after saving or loading
            if phase in {"post_insert", "post_update", "load"} and getattr(
                field_obj, "keep_original", False
            ):
                retval[f"{field_name}_original"] = None

        return retval

    @classmethod
    def validate(cls, kwargs: dict[str, Any]) -> None:
        kwargs.setdefault("secret", True)
        kwargs.setdefault("max_length", 255)
        super().validate(kwargs)


class EmailField(CharField):
    field_type = EmailStr

    @classmethod
    def validate(cls, kwargs: dict[str, Any]) -> None:
        kwargs.setdefault("max_length", 255)
        super().validate(kwargs)


class URLField(CharField):
    @classmethod
    def validate(cls, kwargs: dict[str, Any]) -> None:
        kwargs.setdefault("max_length", 255)
        super().validate(kwargs)


class IPAddressField(FieldFactory, str):
    field_type = Union[ipaddress.IPv4Address, ipaddress.IPv6Address]

    def __new__(  # type: ignore
        cls,
        **kwargs: Any,
    ) -> BaseFieldType:
        kwargs = {
            **kwargs,
            **{key: value for key, value in locals().items() if key not in CLASS_DEFAULTS},
        }

        return super().__new__(cls, **kwargs)

    @classmethod
    def get_column_type(cls, **kwargs: Any) -> IPAddress:
        return IPAddress()

    @staticmethod
    def is_native_type(value: str) -> bool:
        return isinstance(value, (ipaddress.IPv4Address, ipaddress.IPv6Address))

    # overwrite
    @classmethod
    def check(cls, field_obj: BaseFieldType, value: Any, original_fn: Any = None) -> Any:
        if cls.is_native_type(value):
            return value

        match_ipv4 = IPV4_REGEX.match(value)
        match_ipv6 = IPV6_REGEX.match(value)

        if not match_ipv4 and not match_ipv6:  # type: ignore
            raise ValueError("Must be a valid IP format.")

        try:
            return ipaddress.ip_address(value)
        except ValueError:
            raise ValueError("Must be a real IP.")  # noqa



from __future__ import annotations

import asyncio
import warnings
from collections.abc import Iterable
from functools import partial
from inspect import Parameter, Signature, isawaitable
from itertools import islice
from typing import TYPE_CHECKING, Any, Optional, cast

import sqlalchemy

from edgy.core.db.fields.base import BaseField, BaseForeignKey
from edgy.core.db.models.types import BaseModelType
from edgy.core.db.relationships.utils import crawl_relationship

if TYPE_CHECKING:
    from edgy.core.connection.database import Database

    from .types import QuerySetType, tables_and_models_type

generic_field = BaseField()
_forbidden_param_kinds = frozenset([Parameter.KEYWORD_ONLY, Parameter.VAR_KEYWORD])


def is_callable_queryset_filter(inp: Any) -> bool:
    if getattr(inp, "_edgy_force_callable_queryset_filter", None) is not None:
        return cast(bool, inp._edgy_force_callable_queryset_filter)
    if not callable(inp):
        return False
    try:
        signature = Signature.from_callable(inp)
    except ValueError:
        return False
    if len(signature.parameters) < 2:
        return False
    return all(
        param.kind not in _forbidden_param_kinds
        for param in islice(signature.parameters.values(), 2)
    )


async def parse_clause_arg(
    arg: Any, instance: QuerySetType, tables_and_models: tables_and_models_type
) -> Any:
    if is_callable_queryset_filter(arg):
        arg = arg(instance, tables_and_models)
    if isawaitable(arg):
        arg = await arg
    return arg


async def parse_clause_args(
    args: Iterable[Any], queryset: QuerySetType, tables_and_models: tables_and_models_type
) -> list[Any]:
    result: list[Any] = []
    for arg in args:
        result.append(parse_clause_arg(arg, queryset, tables_and_models))
    if queryset.database.force_rollback:
        return [await el for el in result]
    else:
        return await asyncio.gather(*result)


def clean_query_kwargs(
    model_class: type[BaseModelType],
    kwargs: dict[str, Any],
    embed_parent: Optional[tuple[str, str]] = None,
    model_database: Optional[Database] = None,
) -> dict[str, Any]:
    new_kwargs: dict[str, Any] = {}
    for key, val in kwargs.items():
        if embed_parent:
            if embed_parent[1] and key.startswith(embed_parent[1]):
                key = key.removeprefix(embed_parent[1]).removeprefix("__")
            else:
                key = f"{embed_parent[0]}__{key}"
        sub_model_class, field_name, _, _, _, cross_db_remainder = crawl_relationship(
            model_class, key, model_database=model_database
        )
        # we preserve the uncleaned argument
        field = None if cross_db_remainder else sub_model_class.meta.fields.get(field_name)
        if field is not None and not callable(val):
            new_kwargs.update(field.clean(key, val, for_query=True))
        else:
            new_kwargs[key] = val
    assert "pk" not in new_kwargs, "pk should be already parsed"
    return new_kwargs


class _DefaultClausesHelper:
    def __init__(self, op: Any, default_empty: Any) -> None:
        self.op = op
        self.default_empty = default_empty

    def __call__(self, *args: Any) -> Any:
        # unpack, so there are no trues and falses or connectors without any relevance in the where query
        if len(args) == 0:
            return self.default_empty
        if len(args) == 1:
            return args[0]
        return self.op(*args)


def _calculate_select_related(queryset: QuerySetType, *, kwargs: dict[str, Any]) -> set[str]:
    select_related: set[str] = set()
    cleaned_kwargs = clean_query_kwargs(
        queryset.model_class,
        kwargs,
        queryset.embed_parent_filters,
        model_database=queryset.database,
    )
    for key in cleaned_kwargs:
        model_class, field_name, op, related_str, _, cross_db_remainder = crawl_relationship(
            queryset.model_class, key
        )
        if related_str:
            select_related.add(related_str)
    return select_related


def _calculate_select_related_sum(queryset: QuerySetType, *, callables: Iterable[Any]) -> set[str]:
    select_related: set[str] = set()
    for callab in callables:
        select_related.update(callab(queryset))
    return select_related


class _EnhancedClausesHelper:
    def __init__(self, op: Any) -> None:
        self.op = op

    def __call__(self, *args: Any, no_select_related: bool = False) -> Any:
        if all(not is_callable_queryset_filter(arg) and not isawaitable(arg) for arg in args):
            return self.op(*args)
        calculate_select_related_args: list[Any] = (
            []
            if no_select_related
            else [
                arg._edgy_calculate_select_related
                for arg in args
                if hasattr(arg, "_edgy_calculate_select_related")
            ]
        )

        async def wrapper(
            queryset: QuerySetType, tables_and_models: tables_and_models_type
        ) -> Any:
            return self.op(*(await parse_clause_args(args, queryset, tables_and_models)))

        wrapper._edgy_force_callable_queryset_filter = True
        if calculate_select_related_args:
            wrapper._edgy_calculate_select_related = partial(
                _calculate_select_related_sum, callables=calculate_select_related_args
            )
        return wrapper

    def from_kwargs(self, _: Any = None, /, **kwargs: Any) -> Any:
        # ignore first parameter for backward compatibility
        if _ is not None:
            warnings.warn(
                "`from_kwargs` doesn't use the passed positional table or model anymore.",
                DeprecationWarning,
                stacklevel=2,
            )

        async def wrapper(
            queryset: QuerySetType, tables_and_models: tables_and_models_type
        ) -> Any:
            clauses: list[Any] = []
            cleaned_kwargs = clean_query_kwargs(
                queryset.model_class,
                kwargs,
                queryset.embed_parent_filters,
                model_database=queryset.database,
            )

            for key, value in cleaned_kwargs.items():
                model_class, field_name, op, related_str, _, cross_db_remainder = (
                    crawl_relationship(queryset.model_class, key)
                )
                field = model_class.meta.fields.get(field_name, generic_field)
                if cross_db_remainder:
                    assert field is not generic_field
                    fk_field = cast(BaseForeignKey, field)
                    sub_query = (
                        fk_field.target.query.filter(**{cross_db_remainder: value})
                        .only(*fk_field.related_columns.keys())
                        .values_list(fields=fk_field.related_columns.keys())
                    )
                    table = tables_and_models[related_str][0]
                    fk_tuple = sqlalchemy.tuple_(
                        *(getattr(table.columns, colname) for colname in field.get_column_names())
                    )
                    clauses.append(fk_tuple.in_(await sub_query))
                else:
                    assert not isinstance(
                        value, BaseModelType
                    ), f"should be parsed in clean: {key}: {value}"

                    value = await parse_clause_arg(value, queryset, tables_and_models)
                    table = tables_and_models[related_str][0]

                    clauses.append(field.operator_to_clause(field.name, op, table, value))
            return self.op(*clauses)

        wrapper._edgy_force_callable_queryset_filter = True

        if any("__" in key for key in kwargs):
            wrapper._edgy_calculate_select_related = partial(
                _calculate_select_related, kwargs=kwargs
            )

        return wrapper


or_sqlalchemy = _DefaultClausesHelper(sqlalchemy.or_, sqlalchemy.false())
or_sqlalchemy.__doc__ = """
    Creates a SQL Alchemy OR clause for the expressions being passed.
"""
and_sqlalchemy = _DefaultClausesHelper(sqlalchemy.and_, sqlalchemy.true())
and_sqlalchemy.__doc__ = """
    Creates a SQL Alchemy AND clause for the expressions being passed.
"""

or_ = _EnhancedClausesHelper(or_sqlalchemy)
or_.__doc__ = """
    Creates an edgy OR clause for the expressions being passed.
"""
and_ = _EnhancedClausesHelper(and_sqlalchemy)
and_.__doc__ = """
    Creates an edgy AND clause for the expressions being passed.
"""

# alias
Q = and_


def not_(clause: Any, *, no_select_related: bool = False) -> Any:
    """
    Creates a SQL Alchemy NOT clause for the expressions being passed.
    """
    if not is_callable_queryset_filter(clause) and not isawaitable(clause):
        return sqlalchemy.not_(clause)

    async def wrapper(queryset: QuerySetType, tables_and_models: tables_and_models_type) -> Any:
        return sqlalchemy.not_(await parse_clause_arg(clause, queryset, tables_and_models))

    wrapper._edgy_force_callable_queryset_filter = True
    if not no_select_related and hasattr(clause, "_edgy_calculate_select_related"):
        wrapper._edgy_calculate_select_related = clause._edgy_calculate_select_related

    return wrapper


__all__ = [
    "and_",
    "and_sqlalchemy",
    "or_",
    "or_sqlalchemy",
    "not_",
    "generic_field",
    "parse_clause_arg",
    "parse_clause_args",
    "clean_query_kwargs",
]



from __future__ import annotations

from collections.abc import Awaitable, Sequence
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Optional,
)

from pydantic import model_validator
from pydantic.dataclasses import dataclass

if TYPE_CHECKING:
    from edgy.core.db.models.types import BaseModelType

_empty_tuple: tuple[Any, ...] = ()


@dataclass
class Index:
    """
    Class responsible for handling and declaring the database indexes.
    """

    suffix: str = "idx"
    __max_name_length__: ClassVar[int] = 63
    name: Optional[str] = None
    fields: Optional[Sequence[str]] = None

    @model_validator(mode="before")
    def validate_data(cls, values: Any) -> Any:
        name = values.kwargs.get("name")

        if name is not None and len(name) > cls.__max_name_length__:
            raise ValueError(
                f"The max length of the index name must be {cls.__max_name_length__}. Got {len(name)}"
            )

        fields = values.kwargs.get("fields")
        if not isinstance(fields, (tuple, list)):
            raise ValueError("Index.fields must be a list or a tuple.")

        if fields and not all(isinstance(field, str) for field in fields):
            raise ValueError("Index.fields must contain only strings with field names.")

        if name is None:
            suffix = values.kwargs.get("suffix", cls.suffix)
            values.kwargs["name"] = f"{suffix}_{'_'.join(fields)}"
        return values


@dataclass
class UniqueConstraint:
    """
    Class responsible for handling and declaring the database unique_together.
    """

    fields: list[str]
    name: Optional[str] = None
    __max_name_length__: ClassVar[int] = 63

    @model_validator(mode="before")
    def validate_data(cls, values: Any) -> Any:
        name = values.kwargs.get("name")

        if name is not None and len(name) > cls.__max_name_length__:
            raise ValueError(
                f"The max length of the constraint name must be {cls.__max_name_length__}. Got {len(name)}"
            )

        fields = values.kwargs.get("fields")
        if not isinstance(fields, (tuple, list)):
            raise ValueError("UniqueConstraint.fields must be a list or a tuple.")

        if fields and not all(isinstance(field, str) for field in fields):
            raise ValueError("UniqueConstraint.fields must contain only strings with field names.")

        return values


class QueryModelResultCache:
    """
    Class for caching query results.
    """

    def __init__(
        self,
        attrs: Sequence[str],
        prefix: str = "query",
        cache: Optional[dict[str, dict[tuple[Any, ...], Any]]] = None,
    ) -> None:
        if cache is None:
            cache = {}
        self.cache: dict[str, dict[tuple[Any, ...], Any]] = cache
        self.attrs = attrs

    def create_category(
        self, model_class: type[BaseModelType], prefix: Optional[str] = None
    ) -> str:
        return f"{prefix}_{model_class.__name__}"

    def create_sub_cache(
        self, attrs: Sequence[str], prefix: str = "query"
    ) -> QueryModelResultCache:
        return self.__class__(attrs, prefix=prefix, cache=self.cache)

    def clear(
        self, model_class: Optional[type[BaseModelType]] = None, prefix: Optional[str] = None
    ) -> None:
        cache: Any = self.cache
        if model_class is not None:
            cache = cache.get(self.create_category(model_class, prefix=prefix))
        if cache is not None:
            cache.clear()

    def create_cache_key(
        self, instance: Any, prefix: Optional[str] = None, attrs: Optional[Sequence[str]] = None
    ) -> tuple:
        """
        Build a cache key for the model.
        """
        cache_key_list: list[Any] = [self.create_category(type(instance), prefix=prefix)]
        if attrs is None:
            attrs = self.attrs
        # there are no columns, only column results
        if isinstance(instance, dict):
            for attr in attrs:
                cache_key_list.append(str(instance[attr]))
        else:
            for attr in self.attrs:
                cache_key_list.append(str(getattr(instance, attr)))
        return tuple(cache_key_list)

    def get_category(self, model_class: type[BaseModelType], prefix: Optional[str] = None) -> dict:
        return self.cache.setdefault(self.create_category(model_class, prefix=prefix), {})

    def update(self, values: Sequence[Any], cache_keys: Optional[Sequence[tuple]] = None) -> None:
        if cache_keys is None:
            cache_keys = []
            for instance in values:
                try:
                    cache_key = self.create_cache_key(instance)
                except (AttributeError, KeyError):
                    cache_key = _empty_tuple
                cache_keys.append(cache_key)

        for cache_key, instance in zip(cache_keys, values):
            if len(cache_key) <= 1:
                continue
            _category_cache = self.cache.setdefault(cache_key[0], {})
            _cache_list = _category_cache.setdefault(cache_key, [])
            _category_cache[cache_key] = instance

    def get(self, model_class: type[BaseModelType], row_or_model: Any) -> Optional[Any]:
        try:
            cache_key = self.create_cache_key(row_or_model)
        except (AttributeError, KeyError):
            return None
        _category_cache = self.cache.get(model_class.__name__)
        if _category_cache is None:
            return None
        entry = _category_cache.get(cache_key)
        if entry is None:
            return None
        return entry

    def get_or_cache_many(
        self,
        model_class: type[BaseModelType],
        row_or_models: Sequence[Any],
        cache_fn: Optional[Callable[[Any], Optional[BaseModelType]]] = None,
        transform_fn: Optional[Callable[[Optional[BaseModelType]], Any]] = None,
    ) -> Sequence[Any]:
        cache_update_keys: list[tuple] = []
        cache_update: list[BaseModelType] = []
        results: list[Optional[Any]] = []
        for row_or_model in row_or_models:
            result = self.get(model_class, row_or_model)
            if result is None and cache_fn is not None:
                result = cache_fn(row_or_model)
                if result is not None:
                    cache_update_keys.append(self.create_cache_key(result))
                    if transform_fn is not None:
                        result = transform_fn(result)
                    cache_update.append(result)
                results.append(result)
        self.update(cache_update, cache_keys=cache_update_keys)
        return results

    async def aget_or_cache_many(
        self,
        model_class: type[BaseModelType],
        row_or_models: Sequence[Any],
        cache_fn: Optional[Callable[[Any], Awaitable[Optional[BaseModelType]]]] = None,
        transform_fn: Optional[Callable[[Optional[BaseModelType]], Awaitable[Any]]] = None,
    ) -> Sequence[Any]:
        cache_update_keys: list[tuple] = []
        cache_update: list[Any] = []
        results: list[Optional[BaseModelType]] = []
        for row_or_model in row_or_models:
            result = self.get(model_class, row_or_model)
            if result is None:
                result = await cache_fn(row_or_model)
                if result is not None:
                    cache_update_keys.append(self.create_cache_key(result))
                    if transform_fn is not None:
                        result = await transform_fn(result)
                    cache_update.append(result)
                results.append(result)
        self.update(cache_update, cache_keys=cache_update_keys)
        return results



import inspect
import sys
from typing import Any, Callable, NoReturn, Optional, Union

import sqlalchemy
from loguru import logger
from sqlalchemy.sql import schema, sqltypes

import edgy
from edgy import Database, run_sync
from edgy.core.terminal import Print

printer = Print()

SQL_GENERIC_TYPES = {
    sqltypes.BigInteger: edgy.fields.BigIntegerField,
    sqltypes.Integer: edgy.fields.IntegerField,
    sqltypes.JSON: edgy.fields.JSONField,
    sqltypes.Date: edgy.fields.DateField,
    sqltypes.String: edgy.fields.CharField,
    sqltypes.Unicode: edgy.fields.CharField,
    sqltypes.LargeBinary: edgy.fields.BinaryField,
    sqltypes.Boolean: edgy.fields.BooleanField,
    sqltypes.Enum: edgy.fields.ChoiceField,
    sqltypes.DateTime: edgy.fields.DateTimeField,
    sqltypes.Interval: edgy.fields.DurationField,
    sqltypes.Numeric: edgy.fields.DecimalField,
    sqltypes.Float: edgy.fields.FloatField,
    sqltypes.Double: edgy.fields.FloatField,
    sqltypes.SmallInteger: edgy.fields.SmallIntegerField,
    sqltypes.Text: edgy.fields.TextField,
    sqltypes.Time: edgy.fields.TimeField,
    sqltypes.Uuid: edgy.fields.UUIDField,
}

DB_MODULE = "edgy"


def func_accepts_kwargs(func: Callable) -> bool:
    """
    Checks if a function accepts **kwargs.
    """
    return any(
        param
        for param in inspect.signature(func).parameters.values()
        if param.kind == param.VAR_KEYWORD
    )


class InspectDB:
    """
    Class that builds the inspection of a database.
    """

    def __init__(self, database: str, schema: Optional[str] = None) -> None:
        """
        Creates an instance of an InspectDB and triggers the proccess.
        """
        self.database = Database(database)
        self.schema = schema or None

    @staticmethod
    async def reflect(
        *, database: Database, metadata: sqlalchemy.MetaData, **kwargs: Any
    ) -> sqlalchemy.MetaData:
        """
        Connects to the database and reflects all the information about the
        schema bringing all the data available.

        Wrapper around metadata.reflect with logger output
        If you don't want the logger output run:

        `await database.run_sync(metadata.reflect, **kwargs)` directly
        """
        logger.info("Collecting database tables information...")
        await database.run_sync(metadata.reflect, **kwargs)
        return metadata

    async def _inspect(self) -> None:
        async with self.database as database:
            # Connect to a schema
            metadata: sqlalchemy.MetaData = sqlalchemy.MetaData(schema=self.schema)
            metadata = await self.reflect(database=database, metadata=metadata, schema=self.schema)

            # Generate the tables
            tables, _ = self.generate_table_information(metadata)

            for line in self.write_output(tables, str(database.url), schema=self.schema):
                sys.stdout.writelines(line)  # type: ignore

    def inspect(self) -> None:
        """
        Starts the InspectDB and passes all the configurations.
        """
        run_sync(self._inspect())

    @classmethod
    def generate_table_information(
        cls, metadata: sqlalchemy.MetaData
    ) -> tuple[list[dict[str, Any]], dict[str, str]]:
        """
        Generates the tables from the reflection and maps them into the
        `reflected` dictionary of the `Registry`.
        """
        tables_dict = dict(metadata.tables.items())
        tables = []
        models: dict[str, str] = {}
        for key, table in tables_dict.items():
            table_details: dict[str, Any] = {}
            table_details["tablename"] = key.rsplit(".", 1)[-1]

            table_name_list: list[str] = key.split(".")
            table_name = table_name_list[1] if len(table_name_list) > 1 else table_name_list[0]
            table_details["class_name"] = table_name.replace("_", "").replace(".", "").capitalize()
            table_details["class"] = None
            table_details["table"] = table
            models[key] = key.replace("_", "").capitalize()

            # Get the details of the foreign key
            table_details["foreign_keys"] = cls.get_foreign_keys(table)

            # Get the details of the indexes
            table_details["indexes"] = table.indexes
            table_details["constraints"] = table.constraints
            tables.append(table_details)

        return tables, models

    @classmethod
    def get_foreign_keys(
        cls, table_or_column: Union[sqlalchemy.Table, sqlalchemy.Column]
    ) -> list[dict[str, Any]]:
        """
        Extracts all the information needed of the foreign keys.
        """
        details: list[dict[str, Any]] = []

        for foreign_key in table_or_column.foreign_keys:
            fk: dict[str, Any] = {}
            fk["column"] = foreign_key.column
            fk["column_name"] = foreign_key.column.name
            fk["tablename"] = foreign_key.column.table.name
            fk["class_name"] = foreign_key.column.table.name.replace("_", "").capitalize()
            fk["on_delete"] = foreign_key.ondelete
            fk["on_update"] = foreign_key.onupdate
            details.append(fk)

        return details

    @classmethod
    def get_field_type(self, column: sqlalchemy.Column, is_fk: bool = False) -> Any:
        """
        Gets the field type. If the field is a foreign key, this is evaluated,
        outside of the scope.
        """
        if is_fk:
            return "ForeignKey" if not column.unique else "OneToOne", {}

        try:
            real_field: Any = column.type.as_generic()
        except Exception:
            logger.info(
                f"Unable to understand the field type for `{column.type}`, defaulting to TextField."
            )
            real_field = "TextField"

        try:
            field_type = SQL_GENERIC_TYPES[type(real_field)].__name__
        except KeyError:
            logger.info(
                f"Unable to understand the field type for `{column.name}`, defaulting to TextField."
            )
            field_type = "TextField"

        field_params: dict[str, Any] = {}

        if field_type == "CharField":
            field_params["max_length"] = real_field.length

        if field_type in {"CharField", "TextField"} and hasattr(real_field, "collation"):  # noqa: SIM102
            if real_field.collation is not None:
                field_params["collation"] = real_field.collation

        if field_type in {"IntegerField", "SmallIntegerField", "BigIntegerField"}:
            field_params["autoincrement"] = column.autoincrement

        if field_type == "DecimalField":
            field_params["max_digits"] = real_field.precision
            field_params["decimal_places"] = real_field.scale

        if field_type == "FloatField":
            # Note: precision is maybe set to None when reflecting.
            precision = getattr(real_field, "precision", None)
            if precision is None:
                # Oracle
                precision = getattr(real_field, "binary_precision", None)
                if precision is not None:
                    # invert calculation of binary_precision
                    precision = round(precision * 0.30103)
            if precision is not None:
                field_params["max_digits"] = precision

        if field_type == "BinaryField":
            field_params["max_length"] = getattr(real_field, "length", None)

        return field_type, field_params

    @classmethod
    def get_meta(
        cls, table_detail: dict[str, Any], unique_constraints: set[str], _indexes: set[str]
    ) -> NoReturn:
        """
        Produces the Meta class.
        """
        unique_together: list[edgy.UniqueConstraint] = []
        unique_indexes: list[edgy.Index] = []
        indexes = list(table_detail["indexes"])
        constraints = list(table_detail["constraints"])

        # Handle the unique together
        for constraint in constraints:
            if isinstance(constraint, schema.UniqueConstraint) and isinstance(
                constraint.name, str
            ):
                columns = [
                    column.name
                    for column in constraint.columns
                    if column.name not in unique_constraints
                ]
                unique_definition = edgy.UniqueConstraint(fields=columns, name=constraint.name)
                unique_together.append(unique_definition)

        # Handle the indexes
        for index in indexes:
            if isinstance(index, schema.Index):
                columns = [column.name for column in index.columns if column.name not in _indexes]
                index_definition = edgy.Index(name=index.name, fields=columns)
                unique_indexes.append(index_definition)

        meta = [""]
        meta += [
            "    class Meta:\n",
            "        registry = registry\n",
            "        tablename = '{}'\n".format(table_detail["tablename"]),
        ]

        if unique_together:
            meta.append(
                f"        unique_together = {unique_together}\n",
            )

        if unique_indexes:
            meta.append(
                f"        indexes = {unique_indexes}\n",
            )
        return meta

    @classmethod
    def write_output(
        cls, table_details: list[Any], connection_string: str, schema: Union[str, None] = None
    ) -> NoReturn:
        """
        Writes to stdout and runs some internal validations.
        """
        if schema is not None:
            registry = f"registry = {DB_MODULE}.Registry(database=database, schema='{schema}')\n"
        else:
            registry = f"registry = {DB_MODULE}.Registry(database=database)\n"

        yield f"# This is an auto-generated Edgy model module. Edgy version `{edgy.__version__}`.\n"
        yield "#   * Rearrange models' order.\n"
        yield "#   * Make sure each model has one field with primary_key=True.\n"
        yield (
            "#   * Make sure each ForeignKey and OneToOne has `on_delete` set "
            "to the desired behavior.\n"
        )
        yield (
            "# Feel free to rename the models, but don't rename tablename values or "
            "field names.\n"
        )
        yield (
            f"# The generated models do not manage migrations. Those are handled by `{DB_MODULE}.Model`.\n"
        )
        yield f"# The automatic generated models will be subclassed as `{DB_MODULE}.ReflectModel`.\n\n\n"
        yield f"import {DB_MODULE} \n"
        yield f"from {DB_MODULE} import UniqueConstraint, Index \n"

        yield "\n"
        yield "\n"
        yield f"database = {DB_MODULE}.Database('{connection_string}')\n"
        yield registry

        # Start writing the classes
        for table_detail in table_details:
            unique_constraints: set[str] = set()
            indexes: set[str] = set()

            yield "\n"
            yield "\n"
            yield "\n"
            yield "class {}({}.ReflectModel):\n".format(table_detail["class_name"], DB_MODULE)
            # yield "    ...\n"

            sqla_table: sqlalchemy.Table = table_detail["table"]
            if sqla_table.schema:
                yield f'    __using_schema__ = "{sqla_table.schema}"\n'
            columns = list(sqla_table.columns)

            # Get the column information
            for column in columns:
                # ForeignKey related
                foreign_keys = cls.get_foreign_keys(column)
                is_fk: bool = bool(foreign_keys)
                attr_name = column.name

                field_type, field_params = cls.get_field_type(column, is_fk)
                field_params["null"] = column.nullable

                if column.primary_key:
                    field_params["primary_key"] = column.primary_key
                    unique_constraints.add(attr_name)
                if column.unique:
                    unique_constraints.add(attr_name)
                if column.unique and not column.primary_key:
                    field_params["unique"] = column.unique

                if column.index:
                    field_params["index"] = column.index
                    indexes.add(column.name)

                if column.comment:
                    field_params["comment"] = column.comment
                if column.default:
                    field_params["default"] = column.default

                if is_fk:
                    field_params["to"] = foreign_keys[0]["class_name"]
                    field_params["on_update"] = foreign_keys[0]["on_update"]
                    field_params["on_delete"] = foreign_keys[0]["on_update"]
                    field_params["related_name"] = "{}_{}_set".format(
                        attr_name.lower(),
                        field_params["to"].lower(),
                    )

                field_type += "("
                field_description = "{} = {}{}".format(
                    attr_name,
                    "" if "." in field_type else f"{DB_MODULE}.",
                    field_type,
                )
                if field_params:
                    if not field_description.endswith("("):
                        field_description += ", "
                    field_description += ", ".join(f"{k}={v!r}" for k, v in field_params.items())
                field_description += ")\n"
                yield f"    {field_description}"

            yield "\n"
            yield from cls.get_meta(table_detail, unique_constraints, indexes)



from typing import Any

from edgy.core.db.models.metaclasses import (
    BaseModelMeta,
)


class ContentTypeMeta(BaseModelMeta):
    def __new__(
        cls, name: str, bases: tuple[type, ...], attrs: dict[str, Any], **kwargs: Any
    ) -> Any:
        new_model = super().__new__(cls, name, bases, attrs, **kwargs)
        if new_model.no_constraint:
            new_model.__require_model_based_deletion__ = True
        return new_model



from __future__ import annotations

import copy
import inspect
import warnings
from collections.abc import Sequence
from functools import cached_property
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Literal,
    Optional,
    Union,
    cast,
)

import orjson
from pydantic import BaseModel, ConfigDict
from pydantic_core._pydantic_core import SchemaValidator as SchemaValidator

from edgy.core.db.context_vars import (
    CURRENT_INSTANCE,
    CURRENT_MODEL_INSTANCE,
    CURRENT_PHASE,
    MODEL_GETATTR_BEHAVIOR,
)
from edgy.core.db.models.model_reference import ModelRef
from edgy.core.utils.sync import run_sync
from edgy.types import Undefined

from .types import BaseModelType

if TYPE_CHECKING:
    from edgy.core.connection.database import Database
    from edgy.core.db.fields.types import BaseFieldType
    from edgy.core.db.models.model import Model
    from edgy.core.db.querysets.base import QuerySet
    from edgy.core.signals import Broadcaster

_empty = cast(set[str], frozenset())


class EdgyBaseModel(BaseModel, BaseModelType):
    """
    Base of all Edgy models with the core setup.
    """

    model_config = ConfigDict(extra="ignore", arbitrary_types_allowed=True)

    __proxy_model__: ClassVar[Union[type[Model], None]] = None
    __reflected__: ClassVar[bool] = False
    __show_pk__: ClassVar[bool] = False
    __using_schema__: Union[str, None, Any] = Undefined
    # private attribute
    database: ClassVar[Database] = None
    _loaded_or_deleted: bool = False

    def __init__(
        self, *args: Any, __show_pk__: bool = False, __phase__: str = "init", **kwargs: Any
    ) -> None:
        # inject in relation fields anonymous ModelRef (without a Field)
        for arg in args:
            if isinstance(arg, ModelRef):
                relation_field = self.meta.fields[arg.__related_name__]
                extra_params = {}
                try:
                    # m2m or foreign key
                    target_model_class = relation_field.target
                except AttributeError:
                    # reverse m2m or foreign key
                    target_model_class = relation_field.related_from
                if not relation_field.is_m2m:
                    # sometimes the foreign key is required, so set it already
                    extra_params[relation_field.foreign_key.name] = self
                model = target_model_class(
                    **arg.model_dump(exclude={"__related_name__"}),
                    **extra_params,
                )
                existing: Any = kwargs.get(arg.__related_name__)
                if isinstance(existing, Sequence):
                    existing = [*existing, model]
                elif existing is None:
                    existing = [model]
                else:
                    existing = [existing, model]
                kwargs[arg.__related_name__] = existing

        kwargs = self.transform_input(kwargs, phase=__phase__, instance=self)
        super().__init__(**kwargs)
        # restrict to fields
        # TODO: maybe replaceable by direct setting kwargs but doesn't work yet
        self.__dict__ = self.setup_model_from_kwargs(kwargs)
        self.__show_pk__ = __show_pk__
        # always set them in __dict__ to prevent __getattr__ loop
        self._loaded_or_deleted = False

    @classmethod
    def transform_input(
        cls, kwargs: Any, phase: str = "", instance: Optional[BaseModelType] = None
    ) -> Any:
        """
        Expand to_models and apply input modifications.
        """

        kwargs = {**kwargs}
        new_kwargs: dict[str, Any] = {}

        fields = cls.meta.fields
        token = CURRENT_INSTANCE.set(instance)
        token2 = CURRENT_MODEL_INSTANCE.set(instance)
        token3 = CURRENT_PHASE.set(phase)
        try:
            # phase 1: transform
            # Note: this is order dependend. There should be no overlap.
            for field_name in cls.meta.input_modifying_fields:
                fields[field_name].modify_input(field_name, kwargs)
            # phase 2: apply to_model
            for key, value in kwargs.items():
                field = fields.get(key, None)
                if field is not None:
                    new_kwargs.update(**field.to_model(key, value))
                else:
                    new_kwargs[key] = value
        finally:
            CURRENT_PHASE.reset(token3)
            CURRENT_MODEL_INSTANCE.reset(token2)
            CURRENT_INSTANCE.reset(token)
        return new_kwargs

    def setup_model_from_kwargs(self, kwargs: Any) -> Any:
        """
        Loops and setup the kwargs of the model
        """

        return {k: v for k, v in kwargs.items() if k in self.meta.fields}

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {str(self)}>"

    def __str__(self) -> str:
        pkl = []
        token = MODEL_GETATTR_BEHAVIOR.set("passdown")
        try:
            for identifier in self.identifying_db_fields:
                pkl.append(f"{identifier}={getattr(self, identifier, None)}")
        except AttributeError:
            # for abstract embedded
            pass
        finally:
            MODEL_GETATTR_BEHAVIOR.reset(token)
        return f"{self.__class__.__name__}({', '.join(pkl)})"

    @cached_property
    def identifying_db_fields(self) -> Any:
        """The columns used for loading, can be set per instance defaults to pknames"""
        return self.pkcolumns

    @cached_property
    def proxy_model(self) -> type[Model]:
        return self.__class__.proxy_model  # type: ignore

    @property
    def can_load(self) -> bool:
        return bool(
            self.meta.registry
            and not self.meta.abstract
            and all(self.__dict__.get(field) is not None for field in self.identifying_db_fields)
        )

    async def load_recursive(
        self,
        only_needed: bool = False,
        only_needed_nest: bool = False,
        _seen: Optional[set[Any]] = None,
    ) -> None:
        if _seen is None:
            _seen = {self.create_model_key()}
        else:
            model_key = self.create_model_key()
            if model_key in _seen:
                return
            else:
                _seen.add(model_key)
        _loaded_or_deleted = self._loaded_or_deleted
        if self.can_load:
            await self.load(only_needed)
        if only_needed_nest and _loaded_or_deleted:
            return
        for field_name in self.meta.foreign_key_fields:
            value = getattr(self, field_name, None)
            if value is not None:
                # if a subinstance is fully loaded stop
                await value.load_recursive(
                    only_needed=only_needed, only_needed_nest=True, _seen=_seen
                )

    @property
    def signals(self) -> Broadcaster:
        warnings.warn(
            "'signals' has been deprecated, use 'meta.signals' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.meta.signals

    @property
    def fields(self) -> dict[str, BaseFieldType]:
        warnings.warn(
            "'fields' has been deprecated, use 'meta.fields' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.meta.fields

    def model_dump(self, show_pk: Union[bool, None] = None, **kwargs: Any) -> dict[str, Any]:
        """
        An updated version of the model dump.
        It can show the pk always and handles the exclude attribute on fields correctly and
        contains the custom logic for fields with getters

        Extra Args:
            show_pk: bool - Enforces showing the primary key in the model_dump.
        """
        # we want a copy
        exclude: Union[set[str], dict[str, Any], None] = kwargs.pop("exclude", None)
        if exclude is None:
            initial_full_field_exclude = _empty
            # must be writable
            exclude = set()
        elif isinstance(exclude, dict):
            initial_full_field_exclude = {k for k, v in exclude.items() if v is True}
            exclude = copy.copy(exclude)
        else:
            initial_full_field_exclude = set(exclude)
            exclude = copy.copy(initial_full_field_exclude)

        if isinstance(exclude, dict):
            # exclude __show_pk__ attribute from showing up
            exclude["__show_pk__"] = True
            for field_name in self.meta.excluded_fields:
                exclude[field_name] = True
        else:
            exclude.update(self.meta.special_getter_fields)
            exclude.update(self.meta.excluded_fields)
            # exclude __show_pk__ attribute from showing up
            exclude.add("__show_pk__")
        include: Union[set[str], dict[str, Any], None] = kwargs.pop("include", None)
        mode: Union[Literal["json", "python"], str] = kwargs.pop("mode", "python")

        should_show_pk = self.__show_pk__ if show_pk is None else show_pk
        model = super().model_dump(exclude=exclude, include=include, mode=mode, **kwargs)
        # Workaround for metafields, computed field logic introduces many problems
        # so reimplement the logic here
        for field_name in self.meta.special_getter_fields:
            if field_name == "pk":
                continue
            if not should_show_pk or field_name not in self.pknames:
                if field_name in initial_full_field_exclude:
                    continue
                if include is not None and field_name not in include:
                    continue
                if getattr(field_name, "exclude", False):
                    continue
            field: BaseFieldType = self.meta.fields[field_name]
            try:
                retval = field.__get__(self, self.__class__)
            except AttributeError:
                continue
            sub_include = None
            if isinstance(include, dict):
                sub_include = include.get(field_name, None)
                if sub_include is True:
                    sub_include = None
            sub_exclude = None
            if isinstance(exclude, dict):
                sub_exclude = exclude.get(field_name, None)
                if sub_exclude is True:
                    sub_exclude = None
            if isinstance(retval, BaseModel):
                retval = retval.model_dump(
                    include=sub_include, exclude=sub_exclude, mode=mode, **kwargs
                )
            else:
                assert (
                    sub_include is None
                ), "sub include filters for CompositeField specified, but no Pydantic model is set"
                assert (
                    sub_exclude is None
                ), "sub exclude filters for CompositeField specified, but no Pydantic model is set"
                if mode == "json" and not getattr(field, "unsafe_json_serialization", False):
                    # skip field if it isn't a BaseModel and the mode is json and unsafe_json_serialization is not set
                    # currently unsafe_json_serialization exists only on CompositeFields
                    continue
            alias: str = field_name
            if getattr(field, "serialization_alias", None):
                alias = cast(str, field.serialization_alias)
            elif getattr(field, "alias", None):
                alias = field.alias
            model[alias] = retval
        # proxyModel? cause excluded fields to reappear
        # TODO: find a better bugfix
        for excluded_field in self.meta.excluded_fields:
            model.pop(excluded_field, None)
        return model

    def model_dump_json(self, **kwargs: Any) -> str:
        return orjson.dumps(self.model_dump(mode="json", **kwargs)).decode()

    async def execute_pre_save_hooks(
        self, column_values: dict[str, Any], original: dict[str, Any], force_insert: bool
    ) -> dict[str, Any]:
        # also handle defaults
        keys = {*column_values.keys(), *original.keys()}
        affected_fields = self.meta.pre_save_fields.intersection(keys)
        retdict: dict[str, Any] = {}
        if affected_fields:
            # don't trigger loads
            token = MODEL_GETATTR_BEHAVIOR.set("passdown")
            token2 = CURRENT_MODEL_INSTANCE.set(self)
            try:
                for field_name in affected_fields:
                    if field_name not in column_values and field_name not in original:
                        continue
                    field = self.meta.fields[field_name]
                    retdict.update(
                        await field.pre_save_callback(
                            column_values.get(field_name),
                            original.get(field_name),
                            force_insert=force_insert,
                            instance=self,
                        )
                    )
            finally:
                MODEL_GETATTR_BEHAVIOR.reset(token)
                CURRENT_MODEL_INSTANCE.reset(token2)
        return retdict

    async def execute_post_save_hooks(self, fields: Sequence[str], force_insert: bool) -> None:
        affected_fields = self.meta.post_save_fields.intersection(fields)
        if affected_fields:
            # don't trigger loads, AttributeErrors are used for skipping fields
            token = MODEL_GETATTR_BEHAVIOR.set("passdown")
            token2 = CURRENT_MODEL_INSTANCE.set(self)
            try:
                for field_name in affected_fields:
                    field = self.meta.fields[field_name]
                    try:
                        value = getattr(self, field_name)
                    except AttributeError:
                        continue
                    await field.post_save_callback(value, instance=self, force_insert=force_insert)
            finally:
                MODEL_GETATTR_BEHAVIOR.reset(token)
                CURRENT_MODEL_INSTANCE.reset(token2)

    @classmethod
    def extract_column_values(
        cls,
        extracted_values: dict[str, Any],
        is_update: bool = False,
        is_partial: bool = False,
        phase: str = "",
        instance: Optional[Union[BaseModelType, QuerySet]] = None,
        model_instance: Optional[BaseModelType] = None,
    ) -> dict[str, Any]:
        validated: dict[str, Any] = {}
        token = CURRENT_PHASE.set(phase)
        token2 = CURRENT_INSTANCE.set(instance)
        token3 = CURRENT_MODEL_INSTANCE.set(model_instance)
        try:
            # phase 1: transform when required
            if cls.meta.input_modifying_fields:
                extracted_values = {**extracted_values}
                for field_name in cls.meta.input_modifying_fields:
                    cls.meta.fields[field_name].modify_input(field_name, extracted_values)
            # phase 2: validate fields and set defaults for readonly
            need_second_pass: list[BaseFieldType] = []
            for field_name, field in cls.meta.fields.items():
                if field.read_only:
                    # if read_only, updates are not possible anymore
                    if (
                        not is_partial or (field.inject_default_on_partial_update and is_update)
                    ) and field.has_default():
                        validated.update(field.get_default_values(field_name, validated))
                    continue
                if field_name in extracted_values:
                    item = extracted_values[field_name]
                    assert field.owner
                    for sub_name, value in field.clean(field_name, item).items():
                        if sub_name in validated:
                            raise ValueError(f"value set twice for key: {sub_name}")
                        validated[sub_name] = value
                elif (
                    not is_partial or (field.inject_default_on_partial_update and is_update)
                ) and field.has_default():
                    # add field without a value to the second pass (in case no value appears)
                    # only include fields which have inject_default_on_partial_update set or if not is_partial
                    need_second_pass.append(field)

            # phase 3: set defaults for the rest if not partial or inject_default_on_partial_update
            if need_second_pass:
                for field in need_second_pass:
                    # check if field appeared e.g. by composite
                    # Note: default values are directly passed without validation
                    if field.name not in validated:
                        validated.update(field.get_default_values(field.name, validated))
        finally:
            CURRENT_MODEL_INSTANCE.reset(token3)
            CURRENT_INSTANCE.reset(token2)
            CURRENT_PHASE.reset(token)
        return validated

    def __setattr__(self, key: str, value: Any) -> None:
        fields = self.meta.fields
        field = fields.get(key, None)
        token = CURRENT_INSTANCE.set(self)
        token2 = CURRENT_PHASE.set("set")
        try:
            if field is not None:
                if hasattr(field, "__set__"):
                    # not recommended, better to use to_model instead except for kept objects
                    # used in related_fields to mask and not to implement to_model
                    field.__set__(self, value)
                else:
                    for k, v in field.to_model(key, value).items():
                        # bypass __setattr__ method
                        object.__setattr__(self, k, v)
            else:
                # bypass __setattr__ method
                object.__setattr__(self, key, value)
        finally:
            CURRENT_INSTANCE.reset(token)
            CURRENT_PHASE.reset(token2)

    async def _agetattr_helper(self, name: str, getter: Any) -> Any:
        await self.load()
        if getter is not None:
            token = MODEL_GETATTR_BEHAVIOR.set("coro")
            try:
                result = getter(self, self.__class__)
                if inspect.isawaitable(result):
                    result = await result
                return result
            finally:
                MODEL_GETATTR_BEHAVIOR.reset(token)
        try:
            return self.__dict__[name]
        except KeyError:
            raise AttributeError(f"Attribute: {name} not found") from None

    def __getattr__(self, name: str) -> Any:
        """
        Does following things
        1. Initialize managers on access
        2. Redirects get accesses to getter fields
        3. Run an one off query to populate any foreign key making sure
           it runs only once per foreign key avoiding multiple database calls.
        """
        behavior = MODEL_GETATTR_BEHAVIOR.get()
        manager = self.meta.managers.get(name)
        if manager is not None:
            if name not in self.__dict__:
                manager = copy.copy(manager)
                manager.instance = self
                self.__dict__[name] = manager
            return self.__dict__[name]

        field = self.meta.fields.get(name)
        getter: Any = None
        if field is not None and hasattr(field, "__get__"):
            getter = field.__get__
            if behavior == "coro" or behavior == "passdown":
                return field.__get__(self, self.__class__)
            else:
                token = MODEL_GETATTR_BEHAVIOR.set("passdown")
                # no need to set an descriptor object
                try:
                    return field.__get__(self, self.__class__)
                except AttributeError:
                    # forward to load routine
                    pass
                finally:
                    MODEL_GETATTR_BEHAVIOR.reset(token)
        if (
            name not in self.__dict__
            and behavior != "passdown"
            and not self._loaded_or_deleted
            and (field is not None or self.__reflected__)
            and name not in self.identifying_db_fields
            and self.can_load
        ):
            coro = self._agetattr_helper(name, getter)
            if behavior == "coro":
                return coro
            return run_sync(coro)
        return super().__getattr__(name)

    def __eq__(self, other: Any) -> bool:
        # if self.__class__ != other.__class__:
        #     return False
        # somehow meta gets regenerated, so just compare tablename and registry.
        if self.meta.registry is not other.meta.registry:
            return False
        if self.meta.tablename != other.meta.tablename:
            return False
        self_dict = self.extract_column_values(
            self.extract_db_fields(self.pkcolumns),
            is_partial=True,
            phase="compare",
            instance=self,
            model_instance=self,
        )
        other_dict = other.extract_column_values(
            other.extract_db_fields(self.pkcolumns),
            is_partial=True,
            phase="compare",
            instance=other,
            model_instance=other,
        )
        key_set = {*self_dict.keys(), *other_dict.keys()}
        for field_name in key_set:
            if self_dict.get(field_name) != other_dict.get(field_name):
                return False
        return True



from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, cast

from edgy.core.db.models.managers import Manager
from edgy.core.db.models.types import BaseModelType
from edgy.core.db.querysets.clauses import and_

if TYPE_CHECKING:
    from edgy.core.db.querysets.base import QuerySet


class PermissionManager(Manager):
    def permissions_of(self, sources: Sequence[BaseModelType] | BaseModelType) -> QuerySet:
        if isinstance(sources, BaseModelType):
            sources = [sources]
        if len(sources) == 0:
            # none
            return cast("QuerySet", self.filter(and_()))
        UserField = self.owner.meta.fields["users"]
        assert (
            UserField.embed_through is False or UserField.embed_through
        ), "users field need embed_through=foo|False."
        GroupField = self.owner.meta.fields.get("groups", None)
        assert (
            GroupField is None or GroupField.embed_through is False or GroupField.embed_through
        ), "groups field need embed_through=foo|False."

        clauses: list[dict[str, Any]] = []
        for source in sources:
            if isinstance(source, UserField.target):
                clauses.append({"users__pk": source})
                if GroupField is not None:
                    clauses.append({f"groups__{self.owner.users_field_group}__pk": source})
            elif GroupField is not None and isinstance(source, GroupField.target):
                clauses.append({"groups__pk": source})
            else:
                raise ValueError(f"Invalid source: {source}.")
        return cast("QuerySet", self.or_(*clauses))

    def users(
        self,
        permissions: Sequence[str] | str,
        model_names: Sequence[str | None] | str | None = None,
        objects: Sequence[BaseModelType | None] | BaseModelType | None = None,
        include_null_model_name: bool = True,
        include_null_object: bool = True,
    ) -> QuerySet:
        if isinstance(permissions, str):
            permissions = [permissions]
        if isinstance(model_names, str):
            model_names = [model_names]
        if model_names is not None and include_null_model_name:
            model_names = [*model_names, None]
        if isinstance(objects, BaseModelType):
            objects = [objects]
        if objects is not None and include_null_object:
            objects = [*objects, None]
        UserField = self.owner.meta.fields["users"]
        assert (
            UserField.embed_through is False or UserField.embed_through
        ), "users field need embed_through=foo|False."
        GroupField = self.owner.meta.fields.get("groups", None)
        assert (
            GroupField is None or GroupField.embed_through is False or GroupField.embed_through
        ), "groups field need embed_through=foo|False."
        assert (
            GroupField is None
            or GroupField.target.meta.fields[self.owner.users_field_group].embed_through is False
            or GroupField.target.meta.fields[self.owner.users_field_group].embed_through
        ), f"{GroupField.target} {self.owner.users_field_group} field need embed_through=foo|False."
        ModelNameField = self.owner.meta.fields.get("name_model", None)
        ContentTypeField = self.owner.meta.fields.get("obj", None)
        if objects is not None and len(objects) == 0:
            # none
            return cast("QuerySet", UserField.target.query.filter(and_()))
        clauses: list[dict[str, Any]] = [{f"{UserField.reverse_name}__name__in": permissions}]
        if model_names is not None:
            if ModelNameField is not None:
                clauses[-1][f"{UserField.reverse_name}__name_model__in"] = model_names
            elif ContentTypeField is not None:
                clauses[-1][f"{UserField.reverse_name}__obj__name__in"] = model_names

        if GroupField is not None:
            clauses.append({})
            groups_field_user = GroupField.target.meta.fields[
                self.owner.users_field_group
            ].reverse_name
            assert isinstance(
                groups_field_user, str
            ), f"{GroupField.target} {self.owner.users_field_group} field needs reverse_name."
            clauses[-1][f"{groups_field_user}__{GroupField.reverse_name}__name__in"] = permissions
            if model_names is not None:
                if ModelNameField is not None:
                    clauses[-1][
                        f"{groups_field_user}__{GroupField.reverse_name}__name_model__in"
                    ] = model_names
                elif ContentTypeField is not None:
                    clauses[-1][
                        f"{groups_field_user}__{GroupField.reverse_name}__obj__name__in"
                    ] = model_names

        query = cast("QuerySet", UserField.target.query.or_(*clauses))

        if objects is not None:
            obj_clauses = []
            for obj in objects:
                obj_clauses.append({f"{UserField.reverse_name}__obj": obj})
                if GroupField is not None:
                    obj_clauses[-1][
                        f"{self.owner.groups_field_user}__{GroupField.reverse_name}__obj"
                    ] = obj
            query = query.or_(*obj_clauses)
        return query

    def groups(
        self,
        permissions: Sequence[str] | str,
        model_names: Sequence[str | None] | str | None = None,
        objects: Sequence[BaseModelType | None] | BaseModelType | None = None,
        include_null_model_name: bool = True,
        include_null_object: bool = True,
    ) -> QuerySet:
        if isinstance(permissions, str):
            permissions = [permissions]
        if isinstance(model_names, str):
            model_names = [model_names]
        if model_names is not None and include_null_model_name:
            model_names = [*model_names, None]
        if isinstance(objects, BaseModelType):
            objects = [objects]
        if objects is not None and include_null_object:
            objects = [*objects, None]
        GroupField = self.owner.meta.fields["groups"]
        assert (
            GroupField.embed_through is False or GroupField.embed_through
        ), "groups field need embed_through=foo|False."
        assert (
            GroupField.target.meta.fields[self.owner.users_field_group].embed_through is False
            or GroupField.target.meta.fields[self.owner.users_field_group].embed_through
        ), f"{GroupField.target} {self.owner.users_field_group} field need embed_through=foo|False."
        ModelNameField = self.owner.meta.fields.get("name_model", None)
        ContentTypeField = self.owner.meta.fields.get("obj", None)
        if objects is not None and len(objects) == 0:
            # none
            return cast("QuerySet", GroupField.target.query.filter(and_()))
        clauses: dict[str, Any] = {f"{GroupField.reverse_name}__name__in": permissions}
        if model_names is not None:
            if ModelNameField is not None:
                clauses[f"{GroupField.reverse_name}__name_model__in"] = model_names
            elif ContentTypeField is not None:
                clauses[f"{GroupField.reverse_name}__obj__name__in"] = model_names
        query = cast("QuerySet", GroupField.target.query.filter(**clauses))
        if objects is not None:
            for obj in objects:
                clause = {f"{GroupField.reverse_name}__obj": obj}
                query = query.or_(**clause)
        return query



from typing import TYPE_CHECKING, Any, Literal, NamedTuple, Optional, Union

from edgy.core.db.fields.base import BaseForeignKey, RelationshipField

if TYPE_CHECKING:  # pragma: no cover
    from edgy.core.connection.database import Database
    from edgy.core.db.models.types import BaseModelType


class RelationshipCrawlResult(NamedTuple):
    model_class: type["BaseModelType"]
    field_name: str
    operator: Optional[str]
    forward_path: str
    reverse_path: Union[str, Literal[False]]
    cross_db_remainder: str


def crawl_relationship(
    model_class: type["BaseModelType"],
    path: str,
    *,
    model_database: Optional["Database"] = None,
    callback_fn: Any = None,
    traverse_last: bool = False,
) -> RelationshipCrawlResult:
    field = None
    forward_prefix_path = ""
    reverse_path: Union[str, Literal[False]] = ""
    operator: Optional[str] = "exact"
    field_name: str = path
    cross_db_remainder: str = ""
    while path:
        splitted = path.split("__", 1)
        field_name = splitted[0]
        field = model_class.meta.fields.get(field_name)
        if isinstance(field, RelationshipField) and len(splitted) == 2:
            model_class_new, reverse_part, path = field.traverse_field(path)
            if field.is_cross_db(model_database):
                cross_db_remainder = path
                break
            else:
                model_class = model_class_new
                model_database = None
            reverse = not isinstance(field, BaseForeignKey)
            if reverse_part and reverse_path is not False:
                reverse_path = f"{reverse_part}__{reverse_path}" if reverse_path else reverse_part
            else:
                reverse_path = False

            if callback_fn:
                callback_fn(
                    model_class=model_class,
                    field=field,
                    reverse_path=reverse_path,
                    forward_path=forward_prefix_path,
                    reverse=reverse,
                    operator=None,
                    cross_db_remainder=cross_db_remainder,
                )
            if forward_prefix_path:
                forward_prefix_path = f"{forward_prefix_path}__{field_name}"
            else:
                forward_prefix_path = field_name
        elif len(splitted) == 2:
            if "__" not in splitted[1]:
                operator = splitted[1]
                break
            else:
                raise ValueError(
                    f"Tried to cross field: {field_name} of type {field!r}, remainder: {splitted[1]}"
                )
        else:
            operator = "exact"
            break

    if traverse_last and isinstance(field, RelationshipField):
        model_class, reverse_part, path = field.traverse_field(path)
        reverse = not isinstance(field, BaseForeignKey)
    else:
        reverse = False
        reverse_part = field_name
    if reverse_part and reverse_path is not False:
        reverse_path = f"{reverse_part}__{reverse_path}" if reverse_path else reverse_part
    else:
        reverse_path = False
    if callback_fn and field is not None:
        callback_fn(
            model_class=model_class,
            field=field,
            reverse_path=reverse_path,
            forward_path=forward_prefix_path,
            reverse=reverse,
            operator=operator,
            cross_db_remainder=cross_db_remainder,
        )
    return RelationshipCrawlResult(
        model_class=model_class,
        field_name=field_name,
        operator=operator,
        forward_path=forward_prefix_path,
        reverse_path=reverse_path,
        cross_db_remainder=cross_db_remainder,
    )



from typing import ClassVar

import edgy

from .metaclasses import AutoReflectionMeta, AutoReflectionMetaInfo


class AutoReflectModel(edgy.ReflectModel, metaclass=AutoReflectionMeta):
    meta: ClassVar[AutoReflectionMetaInfo]



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Iterable, Sequence
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Optional,
    Union,
)

if TYPE_CHECKING:
    import sqlalchemy
    from databasez.core.transaction import Transaction

    from edgy.core.connection.database import Database
    from edgy.core.db.models.base import BaseModel
    from edgy.core.db.models.managers import BaseManager
    from edgy.core.db.models.metaclasses import MetaInfo
    from edgy.core.db.querysets.base import QuerySet


class DescriptiveMeta:
    """
    The `Meta` class used to configure each metadata of the model.
    Abstract classes are not generated in the database, instead, they are simply used as
    a reference for field generation.

    Usage:

    .. code-block:: python3

        class User(Model):
            ...

            class Meta:
                registry = models
                tablename = "users"

    """

    ...  # pragma: no cover


class BaseModelType(ABC):
    """
    Type of edgy.Model and EdgyBaseModel
    """

    columns: ClassVar[sqlalchemy.sql.ColumnCollection]
    database: ClassVar[Database]
    table: ClassVar[sqlalchemy.Table]
    # Primary key columns
    pkcolumns: ClassVar[Sequence[str]]
    # Primary key fields
    pknames: ClassVar[Sequence[str]]
    query: ClassVar[BaseManager]
    query_related: ClassVar[BaseManager]
    meta: ClassVar[MetaInfo]
    Meta: ClassVar[DescriptiveMeta] = DescriptiveMeta()

    __parent__: ClassVar[Union[type[BaseModelType], None]] = None
    __is_proxy_model__: ClassVar[bool] = False
    __require_model_based_deletion__: ClassVar[bool] = False
    __reflected__: ClassVar[bool] = False
    _db_schemas: ClassVar[dict[str, type[BaseModelType]]]

    @property
    @abstractmethod
    def proxy_model(self) -> type[BaseModelType]: ...

    @property
    @abstractmethod
    def identifying_db_fields(self) -> Any:
        """The columns used for loading, can be set per instance defaults to pknames."""

    @property
    @abstractmethod
    def can_load(self) -> bool:
        """identifying_db_fields are completely specified."""

    @abstractmethod
    def transaction(self, *, force_rollback: bool = False, **kwargs: Any) -> Transaction:
        """Return database transaction for the assigned database."""

    @abstractmethod
    def get_columns_for_name(self, name: str) -> Sequence[sqlalchemy.Column]:
        """Helper for retrieving columns from field name."""

    @abstractmethod
    def identifying_clauses(self) -> Iterable[Any]:
        """Return clauses which are uniquely map to this object"""

    @classmethod
    @abstractmethod
    def generate_proxy_model(cls) -> type[BaseModel]:
        """
        Generates a proxy model for each model. This proxy model is a simple
        shallow copy of the original model being generated.
        """

    @abstractmethod
    async def load(self, only_needed: bool = False) -> None:
        """Load model"""

    @abstractmethod
    async def update(self, **kwargs: Any) -> Any:
        """
        Update operation of the database fields.
        """

    @abstractmethod
    async def save(
        self,
        force_insert: bool = False,
        values: Union[dict[str, Any], set[str], list[str], None] = None,
    ) -> BaseModelType:
        """Save model"""

    @abstractmethod
    async def delete(
        self, skip_post_delete_hooks: bool = False, remove_referenced_call: bool = False
    ) -> None:
        """Delete Model"""

    @abstractmethod
    async def load_recursive(
        self, only_needed: bool = True, only_needed_nest: bool = False
    ) -> None:
        """Load model and all models referenced by foreign keys."""

    @abstractmethod
    def model_dump(self, show_pk: Union[bool, None] = None, **kwargs: Any) -> dict[str, Any]:
        """
        An updated version of the model dump.
        It can show the pk always and handles the exclude attribute on fields correctly and
        contains the custom logic for fields with getters

        Extra Args:
            show_pk: bool - Enforces showing the primary key in the model_dump.
        """

    @classmethod
    @abstractmethod
    def build(
        cls, schema: Optional[str] = None, metadata: Optional[sqlalchemy.MetaData] = None
    ) -> sqlalchemy.Table:
        """
        Builds the SQLAlchemy table representation from the loaded fields.
        """

    @abstractmethod
    async def execute_post_save_hooks(self, fields: Sequence[str], force_insert: bool) -> None: ...

    @abstractmethod
    async def execute_pre_save_hooks(
        self, values: dict[str, Any], original: dict[str, Any], force_insert: bool
    ) -> dict[str, Any]:
        """
        For async operations after clean. Can be used to reintroduce stripped values for save.
        The async operations run in a transaction with save or update. This allows to intervene with the operation.
        Has also access to the defaults and can transform them.

        Returns: column values for saving.
        """

    @classmethod
    @abstractmethod
    def extract_column_values(
        cls,
        extracted_values: dict[str, Any],
        is_update: bool = False,
        is_partial: bool = False,
        instance: Optional[Union[BaseModelType, QuerySet]] = None,
        model_instance: Optional[BaseModelType] = None,
    ) -> dict[str, Any]:
        """
        Extracts all the default values from the given fields and returns the raw
        value corresponding to each field.
        """

    # helpers

    def extract_db_fields(self, only: Optional[Sequence[str]] = None) -> dict[str, Any]:
        """
        Extracts all the db fields, model references and fields.
        Related fields are not included because they are disjoint.
        """
        fields = self.meta.fields
        columns = self.table.columns

        if only is not None:
            return {k: v for k, v in self.__dict__.items() if k in only}

        return {k: v for k, v in self.__dict__.items() if k in fields or hasattr(columns, k)}

    def get_instance_name(self) -> str:
        """
        Returns the name of the class in lowercase.
        """
        return self.__class__.__name__.lower()

    def create_model_key(self) -> tuple:
        """
        Build a cache key for the model.
        """
        pk_key_list: list[Any] = [type(self).__name__]
        # there are no columns, only column results
        for attr in self.pkcolumns:
            pk_key_list.append(str(getattr(self, attr)))
        return tuple(pk_key_list)



from __future__ import annotations

from collections.abc import Awaitable
from contextlib import suppress
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Callable

if TYPE_CHECKING:
    from edgy.core.connection.registry import Registry

ASGIApp = Callable[
    [
        dict[str, Any],
        Callable[[], Awaitable[dict[str, Any]]],
        Callable[[dict[str, Any]], Awaitable[None]],
    ],
    Awaitable[None],
]


class MuteInteruptException(BaseException):
    pass


@dataclass
class ASGIHelper:
    app: ASGIApp
    registry: Registry
    handle_lifespan: bool = False

    async def __call__(
        self,
        scope: dict[str, Any],
        receive: Callable[[], Awaitable[dict[str, Any]]],
        send: Callable[[dict[str, Any]], Awaitable[None]],
    ) -> None:
        if scope["type"] == "lifespan":
            original_receive = receive

            async def receive() -> dict[str, Any]:
                message = await original_receive()
                if message["type"] == "lifespan.startup":
                    try:
                        await self.registry.__aenter__()
                    except Exception as exc:
                        await send({"type": "lifespan.startup.failed", "msg": str(exc)})
                        raise MuteInteruptException from None
                elif message["type"] == "lifespan.shutdown":
                    try:
                        await self.registry.__aexit__()
                    except Exception as exc:
                        await send({"type": "lifespan.shutdown.failed", "msg": str(exc)})
                        raise MuteInteruptException from None
                return message

            if self.handle_lifespan:
                with suppress(MuteInteruptException):
                    while True:
                        message = await receive()
                        if message["type"] == "lifespan.startup":
                            await send({"type": "lifespan.startup.complete"})
                        elif message["type"] == "lifespan.shutdown":
                            await send({"type": "lifespan.shutdown.complete"})
                            break
                return

        with suppress(MuteInteruptException):
            await self.app(scope, receive, send)

    def __getattr__(self, key: str) -> Any:
        # esmerald shim, they extract data directly from app.
        return getattr(self.app, key)



from contextlib import contextmanager
from contextvars import ContextVar, Token
from typing import TYPE_CHECKING, Literal, Optional, Union
from warnings import warn

if TYPE_CHECKING:
    from edgy.core.db.models.types import BaseModelType
    from edgy.core.db.querysets.base import QuerySet

CURRENT_INSTANCE: ContextVar[Optional[Union["BaseModelType", "QuerySet"]]] = ContextVar(
    "CURRENT_INSTANCE", default=None
)
CURRENT_MODEL_INSTANCE: ContextVar[Optional["BaseModelType"]] = ContextVar(
    "CURRENT_MODEL_INSTANCE", default=None
)
CURRENT_PHASE: ContextVar[str] = ContextVar("CURRENT_PHASE", default="")
EXPLICIT_SPECIFIED_VALUES: ContextVar[Optional[set[str]]] = ContextVar(
    "EXPLICIT_SPECIFIED_VALUES", default=None
)
MODEL_GETATTR_BEHAVIOR: ContextVar[Literal["passdown", "load", "coro"]] = ContextVar(
    "MODEL_GETATTR_BEHAVIOR", default="load"
)
TENANT: ContextVar[str] = ContextVar("tenant", default=None)
SCHEMA: ContextVar[str] = ContextVar("SCHEMA", default=None)
# for backward compatibility
SHEMA = SCHEMA


def get_tenant() -> Union[str, None]:
    """
    Gets the current active tenant in the context.
    """
    return TENANT.get()


def set_tenant(value: Union[str, None]) -> Token:
    """
    Sets the global tenant for the context of the queries.
    When a global tenant is set the `get_schema` -> `SCHEMA` is ignored.
    """
    warn(
        "`set_tenant` is deprecated use `with_tenant` instead. WARNING: this function is broken and doesn't reset the tenant.",
        DeprecationWarning,
        stacklevel=2,
    )
    # why deprecating this? It overwrites schema and can lead to hard to debug issues when the scope isn't resetted.

    return TENANT.set(value)


@contextmanager
def with_tenant(tenant: Union[str, None]) -> None:
    """
    Sets the global tenant for the context of the queries.
    When a global tenant is set the `get_schema` -> `SCHEMA` is ignored.
    """
    # Why preferring to a set_tenant?
    # Set_tenant overwrites schema and can lead to hard to debug issues when the scope isn't resetted
    # while this has a limited scope
    # set_tenant affects the get_schema method.
    token = TENANT.set(tenant)
    try:
        yield
    finally:
        TENANT.reset(token)


def get_schema(check_tenant: bool = True) -> Union[str, None]:
    if check_tenant:
        tenant = get_tenant()
        if tenant is not None:
            return tenant
    return SCHEMA.get()


def set_schema(value: Union[str, None]) -> Token:
    """Set the schema and return the token for resetting. Manual way of with_schema."""

    return SCHEMA.set(value)


@contextmanager
def with_schema(schema: Union[str, None]) -> None:
    token = SCHEMA.set(schema)
    try:
        yield
    finally:
        SCHEMA.reset(token)



from typing import TYPE_CHECKING, Optional, Protocol, runtime_checkable

if TYPE_CHECKING:  # pragma: nocover
    from edgy.core.db.models.types import BaseModelType


@runtime_checkable
class ManyRelationProtocol(Protocol):
    instance: "BaseModelType"

    """Defines the what needs to be implemented when using the ManyRelationProtocol"""

    async def save_related(self) -> None: ...

    async def add(self, child: "BaseModelType") -> Optional["BaseModelType"]: ...

    def stage(self, *children: "BaseModelType") -> None:
        """Lazy add children"""

    async def remove(self, child: Optional["BaseModelType"] = None) -> None: ...



from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, Optional, TypeVar, cast

if TYPE_CHECKING:
    from sqlalchemy import Table

    from edgy.core.connection.database import Database
    from edgy.core.connection.registry import Registry
    from edgy.core.db.models.base import BaseModelType
    from edgy.core.db.models.model import Model


def get_model(registry: "Registry", model_name: str) -> "Model":
    """
    Return the model with capitalize model_name.

    Raise lookup error if no model is found.
    """
    return cast("Model", registry.get_model(model_name))


def build_pknames(model_class: Any) -> None:
    """
    Set explicit pknames (field names with primary_key=True set)
    """
    meta = model_class.meta
    pknames: set[str] = set()
    for field_name, field in meta.fields.items():
        if field.primary_key:
            pknames.add(field_name)
    model_class._pknames = tuple(sorted(pknames))


def build_pkcolumns(model_class: Any) -> None:
    """
    Set pkcolumns (columns with primary_key set)
    """
    table = model_class.table
    pkcolumns: set[str] = set()
    for column in table.columns:
        if column.primary_key:
            # key is the sqlalchemy name, in our case name and key should be identically
            pkcolumns.add(column.key)
    model_class._pkcolumns = tuple(sorted(pkcolumns))


def from_model_to_clauses(model: "BaseModelType") -> Iterable[Any]:
    for column in model.columns_for_load:
        yield getattr(model.table.columns, column) == getattr(model, column)


_model_type = TypeVar("_model_type", bound="BaseModelType")


def apply_instance_extras(
    model: _model_type,
    model_class: type["BaseModelType"],
    schema: Optional[str] = None,
    database: Optional["Database"] = None,
    table: Optional["Table"] = None,
) -> _model_type:
    model.__using_schema__ = schema
    if table is None:
        # Apply the schema to the model instance
        model.table = model_class.table_schema(schema)
    else:
        model.table = table
    if database is not None:
        # Set the database to the model instance
        model.database = database
    return model



"""Configuration for Pydantic models."""

from __future__ import annotations as _annotations

from typing_extensions import TypedDict


class ModelConfig(TypedDict, total=False):
    """
    A TypedDict for configuring Model behaviour.
    """

    strict: bool
    """
    Weather the model should be strict in the creation or not.
    """
    populate_by_alias: bool
    """
    Whether an aliased field may be populated by its name as given by the model
    attribute, as well as the alias. Defaults to `False`.
    """
